import logging
import os
from binascii import a2b_base64
import sys
import json
import re
from mimetypes import guess_extension

from six import string_types
import traitlets as traits
from nbconvert.preprocessors import Preprocessor

if sys.version_info[0] == 2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

logger = logging.getLogger("resolve_links")


def guess_extension_without_jpe(mimetype):
    """
    This function fixes a problem with '.jpe' extensions
    of jpeg images which are then not recognised by latex.
    For any other case, the function works in the same way
    as mimetypes.guess_extension
    """
    ext = guess_extension(mimetype)
    if ext == ".jpe":
        ext = ".jpeg"
    return ext


def is_hyperlink(path):
    """test whether a path is a hyperlink, e.g. https://site.org"""
    if urlparse(path).scheme:
        return True
    return False


def resolve_path(fpath, filepath):
    """resolve a relative path, w.r.t. another filepath """

    if is_hyperlink(fpath):
        return fpath

    if not os.path.isabs(fpath):
        fpath = os.path.join(os.path.dirname(str(filepath)), fpath)
        fpath = os.path.abspath(fpath)

    return os.path.normpath(fpath)


def extract_file_links(source, parent_path, redirect_path, replace_nonexistent=False):
    """ extract local linked files

    Examples
    --------
    >>> import os, pytest
    >>> if os.name == 'nt':
    ...     pytest.skip()

    >>> source = '''## Cell with Linked Image
    ... ![test_image](subdir/logo_example.png)
    ... a [test_link](other_doc#a-link)'''
    >>> src, rpaths, npaths = extract_file_links(
    ...                             source, '/root/nb.ipynb', 'redirect', True)
    >>> print(src)
    ## Cell with Linked Image
    ![test_image](redirect/logo_example.png)
    a [test_link](redirect/other_doc#a-link)

    >>> print(rpaths[0])
    /root/subdir/logo_example.png
    >>> print(rpaths[1])
    /root/other_doc


    """
    # TODO is this robust enough
    regex = re.compile("\\[([^\\]]*)\\]\\(([^\\)^\\#]*)([^\\)]*)\\)")
    new_source = source
    redirected_paths = []
    nonexistent_paths = []
    for text, path, label in regex.findall(source):
        if path.startswith("attachment:"):
            continue
        if not path:  # internal links
            continue
        respath = resolve_path(path, parent_path)
        if is_hyperlink(respath):
            continue
        if not os.path.exists(respath):
            nonexistent_paths.append(respath)
        if os.path.exists(respath) or replace_nonexistent:
            redirected_paths.append(respath)
            new_path = os.path.normpath(
                os.path.join(redirect_path, os.path.basename(path))
            )
            new_source = new_source.replace(
                "[{0}]({1}{2})".format(text, path, label),
                "[{0}]({1}{2})".format(text, new_path, label),
            )

    return new_source, redirected_paths, nonexistent_paths


class LatexDocLinks(Preprocessor):
    """ a preprocessor to resolve file paths in the notebook:

    1. Extract attachments from markdown cells, to resources['outputs'],
       and redirect their file links to self.filesfolder

    2. If nb.metadata.ipub.bibliography, create resources['bibliopath']

    3. Creates resources['external_file_paths'] = [] and adds to it:

        - local relative file paths referenced in markdown cells by
          '[](path/to/file)'
        - path to nb.metadata.ipub.bibliography (if present)
        - path to nb.metadata.ipub.titlepage.logo (if present)

    4. If self.redirect_external=True,
       redirects relative external file paths to self.filesfolder

    """

    metapath = traits.Unicode("", help="the file path to the notebook").tag(config=True)
    filesfolder = traits.Unicode(
        "",
        help=("the folder path to dump dump internal content to " "(e.g. images, etc)"),
    ).tag(config=True)
    redirect_external = traits.Bool(
        True, help="if True, redirect relatively linked paths to filesfolder"
    ).tag(config=True)
    extract_attachments = traits.Bool(
        True,
        help=(
            "extract attachments stored in the notebook"
            "(created by dragging and dropping files into markdown cells)"
        ),
    ).tag(config=True)
    output_attachment_template = traits.Unicode(
        "{unique_key}_{cell_index}_{key}{extension}"
    ).tag(config=True)

    def preprocess_cell(self, cell, resources, cell_index):
        """
        Extract attachment

        Parameters
        ----------
        cell : nbformat.notebooknode.NotebookNode
            Notebook cell being processed
        resources : dict
            Additional resources used in the conversion process.  Allows
            preprocessors to pass variables into the Jinja engine.
        cell_index : int
            Index of the cell being processed
        """
        if cell.cell_type != "markdown":
            return cell, resources

        # extract local linked files
        source, rpaths, npaths = extract_file_links(
            cell.source, self.metapath, self.filesfolder
        )
        if self.redirect_external:
            cell.source = source
        resources["external_file_paths"].extend(rpaths)
        resources["unfound_file_paths"].extend(npaths)

        # extract attachments
        unique_key = resources.get("unique_key", "attach")
        if "attachments" in cell and self.extract_attachments:
            attachments = cell.pop("attachments")

            for key, attachment in attachments.items():
                # TODO this only works if there is a single MIME bundle
                (mime_type, data), = attachment.items()

                ext = guess_extension_without_jpe(mime_type)
                if ext is None:
                    ext = "." + mime_type.rsplit("/")[-1]

                # replace the pointer to the attachment
                filepath = os.path.normpath(
                    os.path.join(
                        self.filesfolder,
                        self.output_attachment_template.format(
                            unique_key=unique_key,
                            cell_index=cell_index,
                            key=os.path.splitext(key)[0],
                            extension=ext,
                        ),
                    )
                )
                if "source" in cell:
                    cell["source"] = cell["source"].replace(
                        "attachment:{}".format(key), filepath
                    )

                # code taken from nbconvert.ExtractOutputPreprocessor
                if (
                    not isinstance(data, string_types)
                    or mime_type == "application/json"
                ):
                    # Data is either JSON-like and was parsed into a Python
                    # object according to the spec, or data is for sure
                    # JSON.
                    # In the latter case we want to go extra sure that
                    # we enclose a scalar string value into extra quotes by
                    # serializing it properly.
                    data = json.dumps(data)

                # Binary files are base64-encoded, SVG is already XML
                if mime_type in {"image/png", "image/jpeg", "application/pdf"}:
                    # data is b64-encoded as text (str, unicode),
                    # we want the original bytes
                    data = a2b_base64(data)
                elif sys.platform == "win32":
                    data = data.replace("\n", "\r\n").encode("UTF-8")
                else:
                    data = data.encode("UTF-8")

                if filepath in resources["outputs"]:
                    raise ValueError(
                        "Your outputs have filename metadata associated "
                        "with them. Nbconvert saves these outputs to "
                        "external files using this filename metadata. "
                        "Filenames need to be unique across the notebook, "
                        "or images will be overwritten. The filename {} is"
                        " associated with more than one output. The second"
                        " output associated with this filename is in cell "
                        "{}.".format(filepath, cell_index)
                    )
                # In the resources, make the figure available
                resources["outputs"][filepath] = data

        return cell, resources

    def preprocess(self, nb, resources):
        """
        Preprocessing to apply on each notebook.
        """

        logger.info(
            "resolving external file paths"
            + " in ipub metadata to: {}".format(self.metapath)
        )

        resources.setdefault("external_file_paths", [])
        resources.setdefault("unfound_file_paths", [])

        if "ipub" in nb.metadata:

            if "bibliography" in nb.metadata.ipub:
                bib = nb.metadata.ipub.bibliography
                bib = resolve_path(bib, self.metapath)
                if not os.path.exists(bib):
                    resources["unfound_file_paths"].append(bib)
                else:
                    resources["external_file_paths"].append(bib)
                    resources["bibliopath"] = bib

                if self.redirect_external:
                    nb.metadata.ipub.bibliography = os.path.join(
                        self.filesfolder, os.path.basename(bib)
                    )

            if "filepath" in nb.metadata.ipub.get("bibglossary", {}):
                gloss = nb.metadata.ipub.bibglossary.filepath
                gloss = resolve_path(gloss, self.metapath)
                if not os.path.exists(gloss):
                    resources["unfound_file_paths"].append(gloss)
                else:
                    resources["external_file_paths"].append(gloss)
                    resources["bibglosspath"] = gloss

                if self.redirect_external:
                    nb.metadata.ipub.bibglossary.filepath = os.path.join(
                        self.filesfolder, os.path.basename(gloss)
                    )

            if "logo" in nb.metadata.ipub.get("titlepage", {}):
                logo = nb.metadata.ipub.titlepage.logo
                logo = resolve_path(logo, self.metapath)
                if not os.path.exists(logo):
                    resources["unfound_file_paths"].append(logo)
                else:
                    resources["external_file_paths"].append(logo)

                if self.redirect_external:
                    nb.metadata.ipub.titlepage.logo = os.path.join(
                        self.filesfolder, os.path.basename(logo)
                    )

        for index, cell in enumerate(nb.cells):
            nb.cells[index], resources = self.preprocess_cell(cell, resources, index)

        # filter unique
        resources["external_file_paths"] = list(set(resources["external_file_paths"]))

        upaths = set(resources.pop("unfound_file_paths"))
        if upaths:
            logger.warning(
                "referenced file(s) do not exist" ": {}".format(list(upaths))
            )

        return nb, resources



import logging
import re
import string
import io

import bibtexparser
import traitlets as traits
from nbconvert.preprocessors import Preprocessor

from six import string_types


class DefaultFormatter(string.Formatter):
    def __init__(self, default=""):
        self.default = default

    def get_value(self, key, args, kwds):
        if isinstance(key, string_types):
            return kwds.get(key, self.default.format(key))
        else:
            string.Formatter.get_value(key, args, kwds)


def safe_str(obj):
    if hasattr(obj, "decode"):
        try:
            obj = obj.decode("utf-8")
        except UnicodeEncodeError:
            pass
    try:
        return str(obj)
    except UnicodeEncodeError:
        # python 2.7
        obj = re.sub(u"\u2013", "-", obj)  # en dash
        obj = re.sub(u"\u2014", "--", obj)  # em dash
        return obj.encode("ascii", "ignore").decode("ascii")
    return ""


class LatexTagsToHTML(Preprocessor):
    r""" a preprocessor to find latex tags
    (like ``\cite{abc}`` or ``\todo[color]{stuff}``) and:

    1. attempt to process them into a html friendly format
    2. remove them entirely if this is not possible

    for ``\ref`` or ``\cref``,
    attempts to use resources.refmap to map labels to reference names
    for labels not found in resources.refmap
    the reference name is '<name> <number>', where:

    - <name> is either ref of, if labelbycolon is True and
      the label has a colon, all text before the colon
    - <number> iterate by order of first appearance of a particular label


    NB: should be applied after LatexDocHTML,
    if you want resources.refmap to be available

    Examples
    --------
    >>> from nbformat import NotebookNode
    >>> from jsonextended.utils import MockPath

    >>> processor = LatexTagsToHTML()

    >>> bibfile = MockPath(is_file=True,content='''
    ... @article{bibkey,
    ... title = {the title},
    ... doi = {10.1134/S0018143916050209},
    ... author = {Surname, A. Name},
    ... date = {2016-09-01},
    ... }
    ... ''')
    >>> resources = NotebookNode(
    ...    {'bibliopath':bibfile, 'refmap':{"label":"label_name"}})

    >>> cell = NotebookNode({
    ... "cell_type":"markdown",
    ... "metadata":{},
    ... "source":"test"
    ... })
    >>> nb = NotebookNode({"cells":[cell]})
    >>> nb, _ = processor.preprocess(nb,resources)
    >>> print(nb.cells[0].source)
    test

    >>> cell.source = "\\unknown{test}"
    >>> nb, _ = processor.preprocess(nb,resources)
    >>> print(nb.cells[0].source)
    <BLANKLINE>

    >>> cell.source = "\\ref{label}\\unknown{test}"
    >>> nb, _ = processor.preprocess(nb,resources)
    >>> print(nb.cells[0].source)
    <a href="{id_home_prefix}label">label_name</a>

    >>> cell.source = "\\label{test}"
    >>> nb, _ = processor.preprocess(nb,resources)
    >>> print(nb.cells[0].source)
    <a id="test" class="anchor-link" name="#test">&#182;</a>

    >>> cell.source = "\\cite{bibkey}"
    >>> nb, _ = processor.preprocess(nb,resources)
    >>> print(nb.cells[0].source)
    [<a href="https://doi.org/10.1134/S0018143916050209">Surname <em>et al</em>, 2016.</a>]

    >>> cell.source = "\\begin{equation}x=a+b\\end{equation}"
    >>> nb, _ = processor.preprocess(nb,resources)
    >>> print(nb.cells[0].source)
    \begin{equation}x=a+b\end{equation}

    """  # noqa: E501

    regex = traits.Unicode(
        r"\\(?:[^a-zA-Z]|[a-zA-Z]+[*=']?)(?:\[.*?\])?{.*?}",
        help="the regex to identify latex tags",
    ).tag(config=True)
    bibformat = traits.Unicode(
        "{author}, {year}.",
        help="the format to output \\cite{} tags found in the bibliography",
    ).tag(config=True)
    labelbycolon = traits.Bool(
        True,
        help=(
            "create reference label based on text before colon, "
            "e.g. \\ref{fig:example} -> fig 1"
        ),
    ).tag(config=True)

    def __init__(self, *args, **kwargs):
        # a dictionary to keep track of references,
        # so they each get a different number
        self.refs = {}
        # bibliography references
        self.bibdatabase = {}
        super(LatexTagsToHTML, self).__init__(*args, **kwargs)

    @staticmethod
    def read_bibliography(path):
        """ read a bibliography

        """
        logging.info("reading bibliopath: {}".format(path))
        bibdatabase = {}
        bibparser = bibtexparser.bparser.BibTexParser()
        try:
            if hasattr(path, "open"):
                with path.open(encoding="utf8") as bibtex_file:
                    bibtex_data = bibtex_file.read()
            else:
                with io.open(path, encoding="utf8") as bibtex_file:
                    bibtex_data = bibtex_file.read()
            bibtex_data = safe_str(bibtex_data)
            bibdatabase = bibparser.parse(bibtex_data).entries_dict
        except Exception as err:
            logging.error("could not read bibliopath {}: {}".format(path, err))

        return bibdatabase

    def rreplace(self, source, target, replacement, replacements=1):
        """replace in string, from right-to-left"""
        return replacement.join(source.rsplit(target, replacements))

    def process_bib_entry(self, entry):
        """work out the best way to represent the bib entry """

        # abbreviate a list of authors
        if "author" in entry:
            authors = re.split(", | and ", entry["author"])
            if len(authors) > 1:
                author = authors[0] + " <em>et al</em>"
            else:
                author = authors[0]
            entry["author"] = author

            # split up date into year, month, day
        if "date" in entry:
            date = entry["date"].split("-")
            if len(date) == 3:
                entry["year"] = date[0]
                entry["month"] = date[1]
                entry["day"] = date[2]
            else:
                entry["year"] = date[0]

        text = DefaultFormatter().format(self.bibformat, **entry)

        if "doi" in entry:
            return r'<a href="https://doi.org/{doi}">{text}</a>'.format(
                doi=entry["doi"], text=text
            )
        elif "url" in entry:
            return r'<a href="{url}">{text}</a>'.format(url=entry["url"], text=text)
        elif "link" in entry:
            return r'<a href="{url}">{text}</a>'.format(url=entry["link"], text=text)
        else:
            return text

    def replace_reflabel(self, name, resources):
        """ find a suitable html replacement for a reference label

        the links are left with a format hook in them: {id_home_prefix},
        so that an nbconvert filter can later replace it
        this is particularly useful for slides,
        which require a prefix #/<slide_number><label>
        """
        if "refmap" in resources:
            if name in resources["refmap"]:
                return r'<a href="{{id_home_prefix}}{0}">{1}</a>'.format(
                    name, resources["refmap"][name]
                )

        if self.labelbycolon:
            ref_name = name.split(":")[0] if ":" in name else "ref"
        else:
            ref_name = "ref"
        if ref_name not in self.refs:
            self.refs[ref_name] = {}
        refs = self.refs[ref_name]
        if name in refs:
            id = refs[name]
        else:
            id = len(refs) + 1
            refs[name] = id
        return r'<a href="{{id_home_prefix}}{0}">{1}. {2}</a>'.format(
            name, ref_name, id
        )

    def convert(self, source, resources):
        """ convert a a string with tags in

        Example
        -------

        >>> source = r'''
        ... References to \\cref{fig:example}, \\cref{tbl:example}, \\cref{eqn:example_sympy} and \\cref{code:example_mpl}.
        ...
        ... Referencing multiple items: \\cref{fig:example,fig:example_h,fig:example_v}.
        ...
        ... An unknown latex tag.\\unknown{zelenyak_molecular_2016}
        ... '''
        >>> processor = LatexTagsToHTML()
        >>> print(processor.convert(source,{}))
        <BLANKLINE>
        References to <a href="{id_home_prefix}fig:example">fig. 1</a>, <a href="{id_home_prefix}tbl:example">tbl. 1</a>, <a href="{id_home_prefix}eqn:example_sympy">eqn. 1</a> and <a href="{id_home_prefix}code:example_mpl">code. 1</a>.
        <BLANKLINE>
        Referencing multiple items: <a href="{id_home_prefix}fig:example">fig. 1</a>, <a href="{id_home_prefix}fig:example_h">fig. 2</a> and <a href="{id_home_prefix}fig:example_v">fig. 3</a>.
        <BLANKLINE>
        An unknown latex tag.
        <BLANKLINE>

        """  # noqa: E501
        new = source
        in_equation = False
        labels = []
        for tag in re.findall(self.regex, source):

            if tag.startswith("\\label"):
                link = r'<a id="{label}" class="anchor-link" name="#{label}">&#182;</a>'.format(
                    label=tag[7:-1]
                )  # noqa: E501
                if in_equation:
                    labels.append(link)
                    new = new.replace(tag, "")
                else:
                    new = new.replace(tag, link)

            elif tag.startswith("\\ref"):
                names = tag[5:-1].split(",")
                html = []
                for name in names:
                    html.append(self.replace_reflabel(name, resources))
                new = new.replace(tag, self.rreplace(", ".join(html), ",", " and"))

            elif tag.startswith("\\cref"):
                names = tag[6:-1].split(",")
                html = []
                for name in names:
                    html.append(self.replace_reflabel(name, resources))
                new = new.replace(tag, self.rreplace(", ".join(html), ",", " and"))

            elif tag.startswith("\\cite"):
                names = tag[6:-1].split(",")
                html = []
                for name in names:
                    if name in self.bibdatabase:
                        html.append(self.process_bib_entry(self.bibdatabase[name]))
                    else:
                        html.append("Unresolved citation: {}.".format(name))
                new = new.replace(tag, "[" + ", ".join(html) + "]")

            elif any(
                [
                    tag.startswith("\\begin{{{0}}}".format(env))
                    for env in [
                        "equation",
                        "equation*",
                        "align",
                        "align*",
                        "multline",
                        "multline*",
                        "gather",
                        "gather*",
                    ]
                ]
            ):
                in_equation = True
            elif any(
                [
                    tag.startswith("\\end{{{0}}}".format(env))
                    for env in [
                        "equation",
                        "equation*",
                        "align",
                        "align*",
                        "multline",
                        "multline*",
                        "gather",
                        "gather*",
                    ]
                ]
            ):
                new += " ".join(labels)
                labels = []
                in_equation = False
            elif any(
                [tag.startswith("\\begin{{{0}}}".format(env)) for env in ["split"]]
            ):
                pass
            elif any([tag.startswith("\\end{{{0}}}".format(env)) for env in ["split"]]):
                pass
            else:
                new = new.replace(tag, "")
        return new

    def preprocess(self, nb, resources):

        logging.info("converting latex tags to html")
        if "bibliopath" in resources:
            self.bibdatabase = self.read_bibliography(resources["bibliopath"])
        else:
            self.bibdatabase = {}

        for cell in nb.cells:

            if "ipub" in cell["metadata"]:
                for key in cell["metadata"]["ipub"]:
                    if not isinstance(cell["metadata"]["ipub"][key], dict):
                        continue
                    if "caption" in cell["metadata"]["ipub"][key]:
                        text = cell["metadata"]["ipub"][key]["caption"]
                        key_dict = cell["metadata"]["ipub"][key]
                        key_dict["caption"] = self.convert(text, resources)

            if not cell["cell_type"] == "markdown":
                continue
            cell["source"] = self.convert(cell["source"], resources)

        resources["refslide"] = {}
        return nb, resources



""" a panflute filter to override vs-code Markdown Preview Enhanced (MPE)
interpretation of RMarkdown python cells.

MPE expects cells to have braced attributes after code name;

::
    ```python {cmnd=true}
    print("hi")
    ```

whereas, in RMarkdown, the code name is included in the brace:

::

    ```{python ipub={'figure': {'caption': 'this is a caption'}}}
    print("hi")
    ```

See https://github.com/shd101wyy/vscode-markdown-preview-enhanced/issues/185

"""
# TODO remove when fixed?
import sys
import re
from panflute import Element, Doc, CodeBlock  # noqa: F401
import panflute as pf


def format_code_html(code, doc):
    # type: (CodeBlock, Doc) -> None
    if not (isinstance(code, CodeBlock) and doc.format in ("html", "html5")):
        return None

    if "python" in code.attributes.get("data-info", ""):

        attr = code.attributes.get("data-info", "")

        parsed = "cmd='{}'".format(sys.executable)
        normed = '{{"cmd":"{}"'.format(sys.executable)

        if doc.last_id is None:
            this_id = 1
        else:
            this_id = doc.last_id + 1

        parsed = parsed + " id='{0}'".format(this_id)
        normed = normed + ',"id": "{0}"'.format(this_id)

        match = re.search("\\scontinue=([0-9]+)", attr)
        if match:
            parsed = parsed + " continue='{0}'".format(match.group(1))
            normed = normed + ',"continue":"{0}"'.format(match.group(1))

        if "matplotlib=true" in attr:
            parsed = parsed + " matplotlib=true"
            normed = normed + ',"matplotlib":"true"'

        normed = normed + "}"

        code.attributes["data-info"] = "python {{{0}}}".format(parsed)
        code.attributes[
            "data-parsed-info"
        ] = '{{"language":"python","attributes":{0}}}'.format(normed)
        code.attributes[
            "data-normalized-info"
        ] = '{{"language":"python","attributes":{0}}}'.format(normed)

        doc.last_id = this_id

        return [pf.Para(pf.Str("In [{}]:".format(this_id))), code]


def prepare(doc):
    # type: (Doc) -> None
    doc.last_id = None


def finalize(doc):
    # type: (Doc) -> None
    del doc.last_id


def main(doc=None):
    # type: (Doc) -> None
    return pf.run_filter(format_code_html, prepare, finalize, doc=doc)


if __name__ == "__main__":
    main()



from ipypublish.filters_pandoc.utils import apply_filter
from ipypublish.filters_pandoc import (
    prepare_raw,
    format_cite_elements,
    format_raw_spans,
)


def test_mkdown_hlink_to_rst():
    out_string = apply_filter("[a link](https://pandoc.org/filters.html)", [], "rst")
    assert out_string == "`a link <https://pandoc.org/filters.html>`__"


def test_latex_to_rst():
    """
    """
    in_string = [
        r"\cref{label1} \Cref{label2}  \cite{a-cite-key_2019}",
        "",
        "\\cite{label1,label2}",
        "",
        r"\ref{label3}  \todo{something todo}",
        "",
        r"\todo{something else todo}",
    ]

    out_string = apply_filter(
        in_string,
        [prepare_raw.main, format_cite_elements.main, format_raw_spans.main],
        "rst",
    )

    assert out_string == "\n".join(
        [
            ":ref:`label1` :ref:`label2` :cite:`a-cite-key_2019`",
            "",
            ":cite:`label1,label2`",
            "",
            ":ref:`label3`",
            "",
            ".. todo:: something todo",
            "",
            "",
            "",
            ".. todo:: something else todo",
            "",
            "",
        ]
    )


def test_latex_to_rst_with_numref():
    """"""
    in_string = [
        "---",
        "ipub:",
        "  pandoc:",
        "    use_numref: true",
        "---",
        "",
        r"\cref{label1} \Cref{label2}  \cite{a-cite-key_2019}",
        "",
        r"\ref{label3}  \todo[inline]{something todo}",
        "",
        r"\todo{something else todo}",
    ]

    out_string = apply_filter(
        in_string,
        [prepare_raw.main, format_cite_elements.main, format_raw_spans.main],
        "rst",
    )

    assert out_string.strip() == "\n".join(
        [
            ":numref:`label1` :numref:`label2` :cite:`a-cite-key_2019`",
            "",
            ":ref:`label3`",
            "",
            ".. todo:: something todo" "",
            "",
            "",
            "",
            ".. todo:: something else todo",
        ]
    )


def test_html_to_latex_label():

    in_string = ["[some text](#alabel)"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "latex"
    )

    assert out_string == "\n".join([r"\cref{alabel}"])


def test_cite_in_table_caption():

    in_string = ["a b", "- -", "1 2", "", "Table: Caption \\cite{a}"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "markdown"
    )

    assert out_string == "\n".join(
        ["  a   b", "  --- ---", "  1   2", "", "  : Caption [@a]"]
    )


def test_html_to_latex_cite():

    in_string = [
        'surrounding <cite data-cite="cite_key">text</cite> text' "",
        '<cite data-cite="cite_key2"></cite>',
    ]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "latex"
    )

    assert out_string == "\n".join(
        [r"surrounding \cite{cite_key} text \cite{cite_key2}"]
    )


def test_html_to_rst_cite():

    in_string = [
        'surrounding <cite data-cite="cite_key">text</cite> text',
        "",
        '<cite data-cite="cite_key2"></cite>',
    ]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "rst"
    )

    assert out_string == "\n".join(
        ["surrounding :cite:`cite_key` text", "", ":cite:`cite_key2`"]
    )


def test_citations_latex():

    in_string = [
        "@label1",
        "",
        "[@label1;@label2]",
        "",
        "[an internal link](#label2)" "",
        "[an external link](http://something.org)",
        "",
        "![a citation @label](path/to/image.png)",
    ]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "latex"
    )

    assert out_string.strip() == "\n".join(
        [
            "\\cite{label1}",
            "",
            "\\cite{label1,label2}",
            "",
            "\\cref{label2} \\href{http://something.org}{an external link}",
            "",
            "\\begin{figure}",
            "\\centering",
            "\\includegraphics{path/to/image.png}",
            "\\caption{a citation \\cite{label}}",
            "\\end{figure}",
        ]
    )


def test_citations_rst():

    in_string = [
        "@label1",
        "",
        "[an internal link](#label2)" "",
        "[an external link](http://something.org)",
        "",
        "![a citation @label](path/to/image.png)",
    ]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "rst"
    )

    assert out_string.strip() == "\n".join(
        [
            ":cite:`label1`",
            "",
            ":ref:`label2` `an external link <http://something.org>`__",
            "",
            ".. figure:: path/to/image.png",
            "   :alt: a citation :cite:`label`",
            "",
            "   a citation :cite:`label`",
        ]
    )


def test_rst_cite_to_rst():

    in_string = ["a :ref:`label` b"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "rst"
    )

    assert out_string.strip() == "\n".join(["a :ref:`label` b"])


def test_rst_cite_to_latex():

    in_string = ["a :ref:`label` b"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_cite_elements.main], "latex"
    )

    assert out_string.strip() == "\n".join([r"a \ref{label} b"])


def test_rst_known_role_to_rst():

    in_string = ["a :py:func:`label` b"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_raw_spans.main], "rst"
    )

    assert out_string.strip() == "\n".join(["a :py:func:`label` b"])


def test_rst_directive_to_rst():

    in_string = [".. versionchanged:: v0.8.3", "", "    abc", "", "    xyz"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_raw_spans.main], "rst"
    )

    assert out_string.strip() == "\n".join(
        [".. versionchanged:: v0.8.3", "", "    abc", "", "    xyz"]
    )


def test_rst_directive_to_latex():
    in_string = [".. versionchanged:: v0.8.3", "", "    abc", "", "    xyz"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_raw_spans.main], "latex"
    )

    assert out_string.strip() == "\n".join(
        [
            "\\begin{mdframed}[frametitle={versionchanged},frametitlerule=true]",
            "\\mdfsubtitle{v0.8.3}",
            "",
            "abc",
            "",
            "xyz",
            "",
            "\\end{mdframed}",
        ]
    )


def test_rst_directive_with_options_to_rst():

    in_string = [
        ".. dir::",
        "    :maxdepth: 2",
        "    :numbered:",
        "",
        "    abc",
        "    xyz",
        "",
        "    new paragraph",
        "",
    ]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_raw_spans.main], "rst"
    )

    assert out_string == "\n".join(
        [
            ".. dir::",
            "    :maxdepth: 2",
            "    :numbered:",
            "",
            "    abc",
            "    xyz",
            "",
            "    new paragraph",
            "",
            "",
        ]
    )


def test_rst_label_to_rst():

    in_string = [".. _alabel:"]

    out_string = apply_filter(
        in_string, [prepare_raw.main, format_raw_spans.main], "rst"
    )

    assert out_string.strip() == "\n".join([".. _alabel:"])



# See nbconvert/exporters/html.py:
DISPLAY_DATA_PRIORITY_HTML = (
    "application/vnd.jupyter.widget-state+json",
    "application/vnd.jupyter.widget-view+json",
    "application/javascript",
    "text/html",
    "text/markdown",
    "image/svg+xml",
    "text/latex",
    "image/png",
    "image/jpeg",
    "text/plain",
)
# See nbconvert/exporters/latex.py:
DISPLAY_DATA_PRIORITY_LATEX = (
    "text/latex",
    "image/svg+xml",  # this requires sphinx.ext.imgconverter
    "application/pdf",
    "image/png",
    "image/jpeg",
    "text/markdown",
    "text/plain",
)


def choose_output_type(output, metadata):
    """Choose appropriate output data types for HTML and LaTeX."""
    if output.output_type == "stream":
        html_datatype = latex_datatype = "ansi"
        text = output.text
        output.data = {"ansi": text[:-1] if text.endswith("\n") else text}
    elif output.output_type == "error":
        html_datatype = latex_datatype = "ansi"
        output.data = {"ansi": "\n".join(output.traceback)}
    else:
        for datatype in DISPLAY_DATA_PRIORITY_HTML:
            if datatype in output.data:
                html_datatype = datatype
                break
        else:
            html_datatype = ", ".join(output.data.keys())
        for datatype in DISPLAY_DATA_PRIORITY_LATEX:
            if datatype in output.data:
                latex_datatype = datatype
                break
        else:
            latex_datatype = ", ".join(output.data.keys())

    if html_datatype in [
        "application/vnd.jupyter.widget-state+json",
        "application/vnd.jupyter.widget-view+json",
        "application/javascript",
    ]:
        return html_datatype, latex_datatype

    # Numbering in sphinx fails if splitting outputs into
    # ``.. only:: html`` and ``.. only:: latex``
    # therefore, if possible we should use the same data type for both
    if metadata.ipub.table:
        html_datatype = latex_datatype

    return html_datatype, latex_datatype



""" a panflute filter to find raw elements
and convert them to format agnostic Span elements
"""
import re
from typing import Union  # noqa: F401
from panflute import Element, Doc, Cite, RawInline, Link  # noqa: F401
import panflute as pf

from ipypublish.filters_pandoc.definitions import (
    ATTRIBUTE_CITE_CLASS,
    PREFIX_MAP,
    PREFIX_MAP_LATEX_R,
    PREFIX_MAP_RST_R,
    RST_KNOWN_ROLES,
    RAWSPAN_CLASS,
    RAWDIV_CLASS,
    CONVERTED_CITE_CLASS,
    CONVERTED_OTHER_CLASS,
    CONVERTED_DIRECTIVE_CLASS,
)
from ipypublish.filters_pandoc.utils import get_panflute_containers, get_pf_content_attr


def create_cite_span(identifiers, rawformat, is_block, prefix="", alt=None):
    """create a cite element from an identifier """
    citations = [pf.Citation(identifier) for identifier in identifiers]
    pmapping = dict(dict(PREFIX_MAP)[prefix])
    classes = list(pmapping["classes"])
    classes += [RAWSPAN_CLASS, CONVERTED_CITE_CLASS, ATTRIBUTE_CITE_CLASS]
    attributes = dict(pmapping["attributes"])
    attributes["raw-format"] = rawformat
    if alt is not None:
        attributes["alt"] = str(alt)
    cite = Cite(citations=citations)
    span = pf.Span(cite, classes=classes, attributes=attributes)
    if is_block:
        return pf.Plain(span)
    else:
        return span


def process_internal_links(link, doc):
    # type: (Link, Doc) -> Element
    """extract links that point to internal items, e.g. [text](#label)"""
    if not isinstance(link, pf.Link):
        return None
    match = re.match(r"#(.+)$", link.url)
    if not match:
        return None

    return create_cite_span(
        [match.group(1)],
        "markdown",
        False,
        prefix=dict(PREFIX_MAP_LATEX_R).get("cref"),
        alt=pf.stringify(pf.Plain(*list(link.content))).strip(),
    )


def process_html_cites(container, doc):
    # type: (pf.Block, Doc) -> Element
    """extract raw html <cite data-cite="cite_key">text</cite>"""
    # if not (isinstance(block, get_panflute_containers(pf.RawInline))
    #         or isinstance(block, get_panflute_containers(pf.RawBlock))):
    #     return None
    content_attr = get_pf_content_attr(container, pf.RawInline)
    if not content_attr:
        content_attr = get_pf_content_attr(container, pf.RawBlock)

    if not content_attr:
        return None
    initial_content = getattr(container, content_attr)

    if not initial_content:
        return None

    new_content = []
    skip = 0

    for element in initial_content:

        if skip > 0:
            skip = skip - 1
            continue

        if not (
            isinstance(element, (pf.RawInline, pf.RawBlock))
            and element.format in ("html", "html4", "html5")
        ):
            new_content.append(element)
            continue

        match = re.match(r"<cite\s*data-cite\s*=\"?([^>\"]*)\"?>", element.text)
        if not match:
            new_content.append(element)
            continue

        # look for the closing tag
        span_content = []
        closing = element.next

        while closing:
            if isinstance(closing, pf.RawInline) and closing.format in (
                "html",
                "html5",
            ):
                endmatch = re.match(r"^\s*</cite>\s*$", closing.text)
                if endmatch:
                    break
            span_content.append(closing)
            closing = closing.next

        if not closing:
            new_content.append(element)
            continue

        # TODO include original content
        new_content.append(
            create_cite_span([match.group(1)], "html", isinstance(element, pf.RawBlock))
        )
        skip = len(span_content) + 1

    setattr(container, content_attr, new_content)
    return container


def process_latex_raw(element, doc):
    # type: (Union[pf.RawInline, pf.RawBlock], pf.Doc) -> pf.Element
    """extract all latex adhering to \\tag{content} or \\tag[options]{content}
    to a Span element with class RAWSPAN_CLASS attributes:

    ::

        attributes={"format": "latex",
                    "tag": tag, "content": content, "options": options}

    - Cref, cref, ref, and cite will aslo have class CONVERTED_CITE_CLASS
    - everything else will also have class CONVERTED_OTHER_CLASS

    """
    if not (
        isinstance(element, (pf.RawInline, pf.RawBlock))
        and element.format in ("tex", "latex")
    ):
        return None

    return assess_latex(element.text, isinstance(element, pf.RawBlock))


def process_latex_str(block, doc):
    # type: (pf.Block, Doc) -> Union[pf.Block,None]
    """see process_latex_raw

    same but sometimes pandoc doesn't convert to a raw element
    """
    # TODO why is pandoc sometimes converting latex tags to Str?
    # >> echo "\cite{a}" | pandoc -f markdown -t json
    # {"blocks":[{"t":"Para","c":[{"t":"RawInline","c":["tex","\\cite{a}"]}]}],"pandoc-api-version":[1,17,5,4],"meta":{}}

    content_attr = get_pf_content_attr(block, pf.Str)
    if not content_attr:
        return None
    initial_content = getattr(block, content_attr)

    if not initial_content:
        return None

    new_content = []

    for element in initial_content:
        if not isinstance(element, pf.Str):
            new_content.append(element)
            continue
        for string in re.split(
            r"(\\[^\{\[]+\{[^\}]+\}|\\[^\{\[]+\[[^\]]*\]\{[^\}]+\})", element.text
        ):
            if not string:
                continue
            new_element = assess_latex(string, False)
            if new_element is None:
                new_content.append(pf.Str(string))
            else:
                new_content.append(assess_latex(string, False))

    setattr(block, content_attr, new_content)
    return block


def assess_latex(text, is_block):
    """ test if text is a latex command
    ``\\tag{content}`` or ``\\tag[options]{content}``

    if so return a panflute.Span, with attributes:

    - format: "latex"
    - tag: <tag>
    - options: <options>
    - content: <content>
    - original: <full text>

    """
    # TODO these regexes do not match labels containing nested {} braces
    # use recursive regexes (https://stackoverflow.com/a/26386070/5033292)
    # with https://pypi.org/project/regex/

    # find tags with no option, i.e \tag{label}
    match_latex_noopts = re.match(r"^\s*\\([^\{\[]+)\{([^\}]+)\}\s*$", text)
    if match_latex_noopts:
        tag = match_latex_noopts.group(1)
        content = match_latex_noopts.group(2)
        if tag in dict(PREFIX_MAP_LATEX_R):
            new_element = create_cite_span(
                content.split(","),
                "latex",
                is_block,
                prefix=dict(PREFIX_MAP_LATEX_R).get(tag, ""),
            )
            return new_element

        span = pf.Span(
            classes=[RAWSPAN_CLASS, CONVERTED_OTHER_CLASS],
            attributes={
                "format": "latex",
                "tag": tag,
                "content": content,
                "original": text,
            },
        )
        if is_block:
            return pf.Plain(span)
        else:
            return span

    # find tags with option, i.e \tag[options]{label}
    match_latex_wopts = re.match(r"^\s*\\([^\{\[]+)\[([^\]]*)\]\{([^\}]+)\}\s*$", text)
    if match_latex_wopts:
        tag = match_latex_wopts.group(1)
        options = match_latex_wopts.group(2)
        content = match_latex_wopts.group(3)

        span = pf.Span(
            classes=[RAWSPAN_CLASS, CONVERTED_OTHER_CLASS],
            attributes={
                "format": "latex",
                "tag": tag,
                "content": content,
                "options": options,
                "original": text,
            },
        )
        if is_block:
            return pf.Plain(span)
        else:
            return span

    return None


def process_rst_roles(block, doc):
    # type: (pf.Block, Doc) -> Union[pf.Block,None]
    """extract rst adhering to ``:role:`label```, where role is a known
    to a Cite element with class RAWSPAN_CLASS and CONVERTED_CITE_CLASS
    and attributes:

    ::

        attributes={"format": "rst",
                    "role": tag, "content": content}

    """
    # "a :ref:`label` b" is converted to:
    # (Str(a) Space Str(:ref:) Code(label) Space Str(b))
    # if not (isinstance(block, get_panflute_containers(pf.Str))):
    #     return None
    content_attr = get_pf_content_attr(block, pf.Str)
    if not content_attr:
        return None
    initial_content = getattr(block, content_attr)

    if not initial_content:
        return None

    # match_rst_role = re.match(
    #     "^\\s*\\:([a-z]+)\\:\\`([^\\`]+)\\`$", element.text)

    new_content = []
    skip_next = False

    for element in initial_content:

        if skip_next:
            skip_next = False
            continue

        if not (isinstance(element, pf.Str) and isinstance(element.next, pf.Code)):
            new_content.append(element)
            continue

        if not (
            len(element.text) > 2
            and element.text.startswith(":")
            and element.text.endswith(":")
        ):
            new_content.append(element)
            continue

        role = element.text[1:-1]
        content = element.next.text

        if role in dict(PREFIX_MAP_RST_R):
            new_element = create_cite_span(
                content.split(","),
                "rst",
                False,
                prefix=dict(PREFIX_MAP_RST_R).get(role, ""),
            )
            new_content.append(new_element)
            skip_next = True
        elif role in RST_KNOWN_ROLES:
            new_element = pf.Span(
                classes=[RAWSPAN_CLASS, CONVERTED_OTHER_CLASS],
                attributes={
                    "format": "rst",
                    "role": role,
                    "content": content,
                    "original": "{0}`{1}`".format(element.text, element.next.text),
                },
            )
            new_content.append(new_element)
            skip_next = True
        else:
            new_content.append(element)

    # if len(new_content) != len(block.content):
    #     block.content = new_content
    #     return block
    setattr(block, content_attr, new_content)
    return block


def gather_processors(element, doc):
    """ we gather the processors,
    so that we don't have to do multiple passes
    """

    # apply processors that change one elements

    new_element = process_internal_links(element, doc)
    if new_element is not None:
        return new_element

    new_element = process_latex_raw(element, doc)
    if new_element is not None:
        return new_element

    # apply processors that change multiple inline elements in a block

    if isinstance(element, get_panflute_containers(pf.Inline)) or isinstance(
        pf.Table, pf.DefinitionItem
    ):

        new_element = process_html_cites(element, doc)
        if new_element is not None:
            element = new_element
        new_element = process_latex_str(element, doc)
        if new_element is not None:
            element = new_element
        new_element = process_rst_roles(element, doc)
        if new_element is not None:
            element = new_element

    # apply processors that change multiple block elements
    if isinstance(element, get_panflute_containers(pf.Block)):

        new_element = process_html_cites(element, doc)
        if new_element is not None:
            element = new_element

    return element


def wrap_rst_directives(doc):
    """search for rst directives and wrap them in divs

    with top line starting ``Str(..)Space()Str(name::)``, above a CodeBlock,
    and rst labels of the form ``Str(..)Space()Str(_name:)``

    """
    final_blocks = []
    skip_next = False
    for block in doc.content:

        if skip_next:
            skip_next = False
            continue

        if not isinstance(block, pf.Para):
            final_blocks.append(block)
            continue

        if len(block.content) < 3:
            final_blocks.append(block)
            continue

        if (
            isinstance(block.content[0], pf.Str)
            and block.content[0].text == ".."
            and isinstance(block.content[1], pf.Space)
            and isinstance(block.content[2], pf.Str)
        ):

            if (
                len(block.content) == 3
                and block.content[2].text.startswith("_")
                and block.content[2].text.endswith(":")
            ):
                # the block is an rst label
                new_block = pf.Div(
                    block,
                    classes=[RAWDIV_CLASS, CONVERTED_OTHER_CLASS],
                    attributes={"format": "rst"},
                )
                final_blocks.append(new_block)
                continue

            if block.content[2].text.endswith("::") and isinstance(
                block.next, pf.CodeBlock
            ):
                # the block is a directive with body content
                # TODO at present we allow any directive name
                # the block may contain option directives, e.g. :width:
                skip_next = True

                inline_arg = ""
                if len(block.content) > 3:
                    inline_content = []
                    for el in block.content[3:]:
                        if isinstance(el, pf.SoftBreak):
                            break
                        inline_content.append(el)
                    if inline_content:
                        inline_arg = (
                            pf.stringify(pf.Para(*inline_content))
                            .replace("\n", "")
                            .strip()
                        )

                new_block = pf.Div(
                    block,
                    *pf.convert_text(block.next.text),
                    classes=[RAWDIV_CLASS, CONVERTED_DIRECTIVE_CLASS],
                    attributes={
                        "format": "rst",
                        "directive": block.content[2].text[:-2],
                        "inline": inline_arg,
                        "has_body": True,
                    }
                )
                final_blocks.append(new_block)
                continue

            if block.content[2].text.endswith("::"):
                # the block is a directive without body content
                # TODO at present we allow any directive name
                # the block may contain option directives, e.g. :width:

                inline_arg = ""
                if len(block.content) > 3:
                    inline_content = []
                    for el in block.content[3:]:
                        if isinstance(el, pf.SoftBreak):
                            break
                        inline_content.append(el)
                    if inline_content:
                        inline_arg = (
                            pf.stringify(pf.Para(*inline_content))
                            .replace("\n", "")
                            .strip()
                        )

                new_block = pf.Div(
                    block,
                    classes=[RAWDIV_CLASS, CONVERTED_DIRECTIVE_CLASS],
                    attributes={
                        "format": "rst",
                        "directive": block.content[2].text[:-2],
                        "inline": inline_arg,
                        "has_body": False,
                    },
                )
                final_blocks.append(new_block)
                continue

        final_blocks.append(block)

    doc.content = final_blocks


def prepare(doc):
    # type: (Doc) -> None
    wrap_rst_directives(doc)


def finalize(doc):
    # type: (Doc) -> None
    pass


def main(doc=None, extract_formats=True):
    # type: (Doc, bool) -> None
    """if extract_formats then convert citations defined in
    latex, rst or html formats to special Span elements
    """
    return pf.run_filter(gather_processors, prepare, finalize, doc=doc)


if __name__ == "__main__":
    main()



import pytest


@pytest.mark.requires_latexmk
@pytest.mark.ipynb("nb_with_bib")
def test_withbib_latex(ipynb_app):
    ipynb_app.run(
        {
            "conversion": "latex_ipypublish_main",
            "default_pporder_kwargs": {"create_pdf": True},
        }
    )
    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_exists(extension=".pdf")
    ipynb_app.assert_converted_contains(
        [
            r"\\cite\{zelenyak_molecular_2016\}",
            r"\\bibliographystyle\{unsrtnat\}",
            r"\\bibliography\{main_files/test\}",
        ]
    )


@pytest.mark.requires_latexmk
@pytest.mark.ipynb("nb_with_glossary_tex", out_to_temp=True)
def test_withglossary_tex_latex(ipynb_app):
    ipynb_app.run(
        {
            "conversion": "latex_ipypublish_main",
            "default_pporder_kwargs": {"create_pdf": True},
        }
    )
    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_exists(extension=".pdf")
    ipynb_app.assert_converted_contains(
        [r"\\makeglossaries", r"\\gls\{term1\}", r"\\printglossary"]
    )


@pytest.mark.requires_latexmk
@pytest.mark.ipynb("nb_with_glossary_bib", out_to_temp=True)
def test_withglossary_bib_latex(ipynb_app):
    ipynb_app.run(
        {
            "conversion": "latex_ipypublish_main",
            "default_pporder_kwargs": {"create_pdf": True},
        }
    )
    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_exists(extension=".pdf")
    ipynb_app.assert_converted_contains(
        [r"\\makeglossaries", r"\\gls\{term1\}", r"\\printglossary"]
    )



import docutils.nodes
import docutils.transforms
import sphinx.util
from sphinx import addnodes
from typing import Any  # noqa: F401

from ipypublish.sphinx.gls.bibgloss import format_entries, docutils_citation_node
from ipypublish.sphinx.gls.nodes import BibGlossaryNode

logger = sphinx.util.logging.getLogger(__name__)


class OverrideCitationReferences(docutils.transforms.Transform):
    """
    Replace citation references by pending_xref nodes before the default
    docutils transform tries to resolve them.

    overrides sphinx.transforms.CitationReferences,
    in order to propogate the classes of the citation_reference node
    to the pending_xref node
    """

    # the default priority of sphinx.transforms.CitationReferences is 619
    default_priority = 618

    def apply(self, **kwargs):
        # type: (Any) -> None
        # mark citation labels as not smartquoted
        # for citation in self.document.traverse(nodes.citation):
        #     label = cast(nodes.label, citation[0])
        #     label['support_smartquotes'] = False

        for citation_ref in self.document.traverse(docutils.nodes.citation_reference):
            cittext = citation_ref.astext()
            refnode = addnodes.pending_xref(
                cittext,
                refdomain="std",
                reftype="citation",
                reftarget=cittext,
                refwarn=True,
                support_smartquotes=False,
                ids=citation_ref["ids"],
            )
            refnode.source = citation_ref.source or citation_ref.parent.source
            refnode.line = citation_ref.line or citation_ref.parent.line
            refnode += docutils.nodes.Text("[" + cittext + "]")
            for class_name in citation_ref.attributes.get("classes", []):
                refnode["classes"].append(class_name)
            citation_ref.parent.replace(citation_ref, refnode)


class HandleMissingCitesTransform(docutils.transforms.Transform):
    """ before sphinx.transforms.post_transforms.ReferencesResolver
    missing citations need to be handled (default_priority=10)
    """

    default_priority = 9

    def apply(self):
        # type: () -> None
        app = self.document.settings.env.app
        for node in self.document.traverse(addnodes.pending_xref):
            if "bibglossary" not in node.attributes.get("classes", []):
                continue
            text = node[0].astext()
            key = text[1:-1]
            try:
                app.env.bibgloss_cache.get_label_from_key(key)
            except KeyError:
                logger.warning(
                    "could not relabel bibglossary reference [%s]" % key,
                    type="bibgloss",
                    subtype="relabel",
                )
                # strip class (otherwise ReferencesResolver fails)
                node.attributes["classes"] = []


def node_text_transform(node, transform):
    """Apply transformation to all Text nodes within node."""
    for child in node.children:
        if isinstance(child, docutils.nodes.Text):
            node.replace(child, transform(child))
        else:
            node_text_transform(child, transform)


def transform_url_command(textnode):
    """Convert '\\\\url{...}' into a proper docutils hyperlink."""
    text = textnode.astext()
    if "\\url" in text:
        text1, _, text = text.partition("\\url")
        text2, _, text3 = text.partition("}")
        text2 = text2.lstrip(" {")
        ref = docutils.nodes.reference(refuri=text2)
        ref += docutils.nodes.Text(text2)
        node = docutils.nodes.inline()
        node += transform_url_command(docutils.nodes.Text(text1))
        node += ref
        node += transform_url_command(docutils.nodes.Text(text3))
        return node
    else:
        return textnode


class BibGlossaryTransform(docutils.transforms.Transform):
    """A docutils transform to generate citation entries for
    bibglossary nodes.
    """

    # transform must be applied before references are resolved
    default_priority = 10
    """Priority of the transform. See
    http://docutils.sourceforge.net/docs/ref/transforms.html
    """

    def apply(self):
        """Transform each
        :class:`~ipypublish.sphinx.gls.nodes.BibGlossaryNode`
        node into a list of glossary terms.
        """
        env = self.document.settings.env
        docname = env.docname
        for bibnode in self.document.traverse(BibGlossaryNode):

            id_ = bibnode["ids"][0]
            bibcache = env.bibgloss_cache.get_bibliography_cache(
                docname=docname, id_=id_
            )
            entries = env.bibgloss_cache.get_bibliography_entries(
                docname=docname, id_=id_, warn=logger.warning
            )

            # create citation nodes for all references
            nodes = docutils.nodes.paragraph()

            for styled_entry in format_entries(
                entries, style=bibcache.style, sort=not bibcache.unsorted
            ):

                citation = docutils_citation_node(
                    styled_entry,
                    self.document,
                    convert_latex=env.app.config.bibgloss_convert_latex,
                )
                # docutils_citation_node(...) uses entry.key
                # as citation label
                # we change it to entry.label later onwards
                # but we must note the entry.label now;
                # at this point, we also already prefix the label
                key = citation[0].astext()
                bibcache.labels[key] = styled_entry.label
                bibcache.plurals[key] = styled_entry.plural

                node_text_transform(citation, transform_url_command)
                nodes += citation

            bibnode.replace_self(nodes)



print("import: 'ipypublish'")
import ipypublish

print("import: 'ipypublish.bib2glossary'")
import ipypublish.bib2glossary

print("import: 'ipypublish.convert'")
import ipypublish.convert

print("import: 'ipypublish.export_plugins'")
import ipypublish.export_plugins

print("import: 'ipypublish.filters'")
import ipypublish.filters

print("import: 'ipypublish.filters_pandoc'")
import ipypublish.filters_pandoc

print("import: 'ipypublish.filters_pandoc.tests'")
import ipypublish.filters_pandoc.tests

print("import: 'ipypublish.frontend'")
import ipypublish.frontend

print("import: 'ipypublish.port_api'")
import ipypublish.port_api

print("import: 'ipypublish.postprocessors'")
import ipypublish.postprocessors

print("import: 'ipypublish.preprocessors'")
import ipypublish.preprocessors

print("import: 'ipypublish.schema'")
import ipypublish.schema

print("import: 'ipypublish.scripts'")
import ipypublish.scripts

print("import: 'ipypublish.sphinx'")
import ipypublish.sphinx

print("import: 'ipypublish.sphinx.gls'")
import ipypublish.sphinx.gls

print("import: 'ipypublish.sphinx.notebook'")
import ipypublish.sphinx.notebook

print("import: 'ipypublish.sphinx.notebook.css'")
import ipypublish.sphinx.notebook.css

print("import: 'ipypublish.sphinx.tests'")
import ipypublish.sphinx.tests

print("import: 'ipypublish.templates'")
import ipypublish.templates

print("import: 'ipypublish.templates.outline_schemas'")
import ipypublish.templates.outline_schemas

print("import: 'ipypublish.templates.segments'")
import ipypublish.templates.segments

print("import: 'ipypublish.tests'")
import ipypublish.tests

print("import: 'ipypublish.tests.test_files'")
import ipypublish.tests.test_files




import os
import pytest

from ipypublish.postprocessors.pdfexport import PDFExport
from ipypublish.postprocessors.reveal_serve import RevealServer


@pytest.mark.requires_latexmk
def test_pdf_export(temp_folder):

    tex_content = """
\\documentclass{article}
\\begin{document}
hallo world
\\end{document}
"""

    tex_path = os.path.join(temp_folder, "test.tex")
    pdf_path = os.path.join(temp_folder, "test.pdf")

    with open(tex_path, "w") as f:
        f.write(tex_content)

    pdfexport = PDFExport()
    pdfexport.postprocess(tex_content, "text/latex", tex_path)

    assert os.path.exists(pdf_path)


def test_reveal_server():
    RevealServer()
    # TODO test reveal server runs correctly,
    # possibly use https://github.com/eugeniy/pytest-tornado



import sys

import pytest


@pytest.mark.skipif(
    sys.version_info[:2] == (3, 5), reason="json_dumps puts keys in different order"
)
@pytest.mark.ipynb("nb_with_ipywidget")  # , out_to_temp=False)
def test_ipywidget_sphinx_rst(ipynb_app):
    """The notebook contains an ipywidgets and the widget state has been saved."""

    ipynb_app.run({"conversion": "sphinx_ipypublish_all"})

    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_equals_expected("main")



""" a panflute filter to format Span element
representations of RawInline elements

The :py:mod:`ipypublish.filters_pandoc.prepare_raw` filter should be run
first to access the functionality below:

"""
import itertools

# from textwrap import fill as textwrap

from panflute import Element, Doc, Span  # noqa: F401
import panflute as pf

from ipypublish.filters_pandoc.definitions import (
    CONVERTED_OTHER_CLASS,
    CONVERTED_DIRECTIVE_CLASS,
    IPUB_META_ROUTE,
)


def process_raw_spans(container, doc):
    # type: (Span, Doc) -> Element
    if not isinstance(container, (pf.Span, pf.Div)):
        return None

    hide_raw = doc.get_metadata(IPUB_META_ROUTE + ".hide_raw", False)

    if CONVERTED_OTHER_CLASS in container.classes and isinstance(container, pf.Span):
        if doc.format == "rst" and container.attributes["format"] == "latex":
            if container.attributes["tag"] in ["todo"]:
                return pf.Str(
                    "\n\n.. {}:: {}\n\n".format(
                        container.attributes["tag"], container.attributes["content"]
                    )
                )
            if container.attributes["tag"] == "ensuremath":
                return pf.RawInline(
                    ":math:`{}`".format(container.attributes["content"]), format="rst"
                )

        return pf.RawInline(
            container.attributes.get("original"), format=container.attributes["format"]
        )

    if CONVERTED_DIRECTIVE_CLASS in container.classes and isinstance(container, pf.Div):
        # convert the directive head, which will be e.g.
        # Para(Str(..) Space Str(toctree::) SoftBreak Str(:maxdepth:) Space Str(2) SoftBreak Str(:numbered:))  # noqa
        # we need to spilt on the soft breaks,
        # place them on a new line and re-indent them

        if doc.format in ("rst"):

            # split into lines by soft breaks
            header_lines = [
                list(y)
                for x, y in itertools.groupby(
                    container.content[0].content, lambda z: isinstance(z, pf.SoftBreak)
                )
                if not x
            ]

            # wrap each line in a Para and convert block with pandoc
            head_doc = pf.Doc(*[pf.Para(*l) for l in header_lines])
            head_doc.api_version = doc.api_version
            head_str = pf.convert_text(
                head_doc, input_format="panflute", output_format=doc.format
            )
            # remove blank lines and indent
            head_str = head_str.replace("\n\n", "\n    ") + "\n\n"
            head_block = pf.RawBlock(head_str, format=doc.format)

            if len(container.content) == 1:
                return head_block

            # split into lines by soft breaks, we use indicators to tell
            # us where to indent in the converted text
            body_blocks = []
            for block in container.content[1:]:
                new_elements = [pf.RawInline("%^*", format=doc.format)]
                for el in block.content:
                    if isinstance(el, pf.SoftBreak):
                        new_elements.append(pf.RawInline("?&@", format=doc.format))
                    else:
                        new_elements.append(el)
                block.content = new_elements
                body_blocks.append(block)

            # convert body content with pandoc
            body_doc = pf.Doc(*body_blocks)
            body_doc.api_version = doc.api_version
            body_str = pf.convert_text(
                body_doc, input_format="panflute", output_format=doc.format
            )
            # raise ValueError(body_blocks)
            body_str = body_str.replace("%^*", "    ").replace("?&@", "\n    ")

            # ensure all lines are indented correctly
            # (doesn't occur by default?)
            body_str = (
                "\n".join(
                    [
                        "    " + l.lstrip() if l.strip() else l
                        for l in body_str.splitlines()
                    ]
                )
                + "\n\n"
            )

            body_block = pf.RawBlock(body_str, format=doc.format)
            return [head_block, body_block]

        elif (
            doc.format in ("html", "html5") and container.attributes["format"] == "rst"
        ):

            if hide_raw:
                return []

            head_para = pf.Para(
                *[
                    pf.RawInline("<br>" + "&nbsp" * 4)
                    if isinstance(c, pf.SoftBreak)
                    else c
                    for c in container.content[0].content
                ]
            )
            head_str = pf.convert_text(
                head_para, input_format="panflute", output_format=doc.format
            )

            if len(container.content) > 1:

                body_doc = pf.Doc(*container.content[1:])
                body_doc.api_version = doc.api_version
                body_str = pf.convert_text(
                    body_doc, input_format="panflute", output_format=doc.format
                )
                body_str = (
                    '<p></p><div style="margin-left: 20px">' "{0}</div>"
                ).format(body_str)
            else:
                body_str = ""

            return pf.RawBlock(
                '<div {0} style="background-color:rgba(10, 225, 10, .2)">'
                "{1}{2}"
                "</div>".format(
                    container.attributes.get("directive", ""), head_str, body_str
                ),
                format="html",
            )

        elif doc.format in ("tex", "latex") and container.attributes["format"] == "rst":

            if hide_raw:
                return []

            directive = container.attributes.get("directive", "")
            inline = container.attributes.get("inline", "")
            # TODO handle directive with options and/or inline body
            # e.g. .. figure:: path/to/figure
            #          :centre:

            box_open = (
                "\\begin{{mdframed}}"
                "[frametitle={{{0}}},frametitlerule=true]".format(directive)
            )
            if inline:
                box_open += "\n\\mdfsubtitle{{{0}}}".format(inline)
            box_close = "\\end{mdframed}"

            if len(container.content) == 1:
                return pf.RawBlock(box_open + box_close, format="tex")
            else:
                return (
                    [pf.RawBlock(box_open, format="tex")]
                    + list(container.content[1:])
                    + [pf.RawBlock(box_close, format="tex")]
                )

        return pf.RawBlock(
            pf.stringify(pf.Doc(*container.content)),
            format=container.attributes["format"],
        )

    if CONVERTED_OTHER_CLASS in container.classes and isinstance(container, pf.Div):
        return pf.RawBlock(
            pf.stringify(pf.Doc(*container.content)),
            format=container.attributes["format"],
        )


# now unused
# def split_soft_breaks(container,
#                       indent=4, fmt="rst", indent_first=False,
#                       pre_content="", post_content="",
#                       pre_chunk="", post_chunk="",
#                       linebreak="\n", raw_indent=None):
#     """rst conversion doesn't recognise soft breaks as new lines,
#     so add them manually and return a list containing the new elements
#     """
#     content = []
#     if pre_content:
#         content.append(pf.RawBlock(pre_content, fmt))

#     chunks = [list(y) for x, y in itertools.groupby(
#         container.content,
#         lambda z: isinstance(z, pf.SoftBreak)) if not x]

#     for i, chunk in enumerate(chunks):
#         if i > 0 or indent_first:
#             if raw_indent is not None:
#                 chunk = [pf.RawInline(raw_indent, fmt)] * indent + chunk
#             else:
#                 chunk = [pf.Space()] * indent + chunk

#         if pre_chunk:
#             content.append(pf.RawBlock(pre_chunk, fmt))
#         content.append(pf.Plain(*chunk))
#         content.append(pf.RawBlock(linebreak, fmt))
#         if post_chunk:
#             content.append(pf.RawBlock(post_chunk, fmt))

#     # if isinstance(container, pf.Para):
#     #     content.append(pf.RawBlock(linebreak, fmt))

#     if post_content:
#         content.append(pf.RawBlock(post_content, fmt))

#     return content


def process_code_latex(code, doc):
    # type: (pf.CodeBlock, Doc) -> Element
    if doc.format not in ("tex", "latex"):
        return None
    if not isinstance(code, pf.CodeBlock):
        return None
    # TODO line wrapping
    return [
        pf.RawBlock("\\begin{mdframed}", format=doc.format),
        code,
        pf.RawBlock("\\end{mdframed}", format=doc.format),
    ]


def prepare(doc):
    # type: (Doc) -> None
    pass


def finalize(doc):
    # type: (Doc) -> None
    pass


def main(doc=None):
    # type: (Doc) -> None
    """
    """
    return pf.run_filters(
        [process_raw_spans, process_code_latex], prepare, finalize, doc=doc
    )


if __name__ == "__main__":
    main()



import logging

import traitlets as traits
from nbconvert.preprocessors import Preprocessor


class CropCells(Preprocessor):
    """A preprocessor to crop the notebook cells from <start> to <end>"""

    start = traits.Integer(0, help="first cell of notebook to be converted").tag(
        config=True
    )
    end = traits.Integer(-1, help="last cell of notebook to be converted").tag(
        config=True
    )

    def preprocess(self, nb, resources):
        logging.info(
            "preprocessing notebook: cropping cells {0} to {1}".format(
                self.start, self.end
            )
        )
        nb.cells = nb.cells[self.start : self.end]
        return nb, resources



"""
Originally adapted from: https://github.com/mcmtroffaes/sphinxcontrib-bibtex
"""
import docutils
import sphinx.util

from ipypublish.sphinx.gls.cache import Cache
from ipypublish.sphinx.gls.bibgloss import latex_to_docutils

logger = sphinx.util.logging.getLogger(__name__)


def init_bibgloss_cache(app):
    """Create ``app.env.bibgloss_cache`` if it does not exist yet.
    :param app: The sphinx application.
    :type app: :class:`sphinx.application.Sphinx`
    """
    if not hasattr(app.env, "bibgloss_cache"):
        app.env.bibgloss_cache = Cache()


def purge_bibgloss_cache(app, env, docname):
    """Remove all information related to *docname* from the cache.
    :param app: The sphinx application.
    :type app: :class:`sphinx.application.Sphinx`
    :param env: The sphinx build environment.
    :type env: :class:`sphinx.environment.BuildEnvironment`
    """
    env.bibgloss_cache.purge(docname)


def process_citations(app, doctree, docname):
    """Replace labels of citation nodes by actual labels.
    :param app: The sphinx application.
    :type app: :class:`sphinx.application.Sphinx`
    :param doctree: The document tree.
    :type doctree: :class:`docutils.nodes.document`
    :param docname: The document name.
    :type docname: :class:`str`
    """
    # citation(rawsource='', *children, **attributes)
    for node in doctree.traverse(docutils.nodes.citation):
        if "bibglossary" not in node.attributes.get("classes", []):
            continue
        key = node[0].astext()
        try:
            label_str = app.env.bibgloss_cache.get_label_from_key(key)
        except KeyError:
            logger.warning(
                "could not relabel glossary item [%s]" % key,
                type="bibgloss",
                subtype="relabel",
            )
        else:
            if app.config.bibgloss_convert_latex:
                label = latex_to_docutils(label_str)
                # label(rawsource='', text='', *children, **attributes)
                node[0] = docutils.nodes.label("", "", *label.children[0])
            else:
                node[0] = docutils.nodes.label("", label_str)

            if sphinx.version_info >= (2,):
                # this creates a line break between the label and the definition
                node.insert(1, docutils.nodes.line())


def process_citation_references(app, doctree, docname):
    """Replace text of citation reference nodes by actual labels.
    :param app: The sphinx application.
    :type app: :class:`sphinx.application.Sphinx`
    :param doctree: The document tree.
    :type doctree: :class:`docutils.nodes.document`
    :param docname: The document name.
    :type docname: :class:`str`
    """
    # sphinx has already turned citation_reference nodes
    # into reference nodes, so iterate over reference nodes
    # reference(rawsource='', text='', *children, **attributes)
    for node in doctree.traverse(docutils.nodes.reference):
        if "bibglossary" not in node.attributes.get("classes", []):
            continue
        text = node[0].astext()
        key = text[1:-1]
        try:
            if "bibgplural" in node.attributes.get("classes", []):
                label = app.env.bibgloss_cache.get_plural_from_key(key)
            else:
                label = app.env.bibgloss_cache.get_label_from_key(key)
        except KeyError:
            pass
            logger.warning(
                "could not relabel glossary reference [%s]" % key,
                type="bibgloss",
                subtype="relabel",
            )
        else:
            if "bibgcapital" in node.attributes.get("classes", []):
                label = label.capitalize()

            if app.config.bibgloss_convert_latex:
                # note we use only the first child of the document
                # TODO is this the best way to do this
                label = latex_to_docutils(label)
                if sphinx.version_info < (2,):
                    node[0] = label.children[0]
                else:
                    # in sphinx v2 the child is a paragraph
                    node[0] = label.children[0].children[0]
            else:
                node[0] = docutils.nodes.Text(label)



IPUB_META_ROUTE = "ipub.pandoc"

ATTRIBUTE_CITE_CLASS = "attribute-Cite"
RAWSPAN_CLASS = "raw-span-content"
RAWDIV_CLASS = "raw-div-content"
CONVERTED_CITE_CLASS = "converted-Cite"
CONVERTED_DIRECTIVE_CLASS = "converted-rst-dir"
CONVERTED_OTHER_CLASS = "converted-Other"

# NB: it appears '-' and '^' are already used by sphinx
PREFIX_MAP = (
    ("", (("classes", ()), ("attributes", (("latex", "cite"), ("rst", "cite"))))),
    ("+", (("classes", ()), ("attributes", (("latex", "cref"), ("rst", "numref"))))),
    ("!", (("classes", ()), ("attributes", (("latex", "ref"), ("rst", "ref"))))),
    ("=", (("classes", ()), ("attributes", (("latex", "eqref"), ("rst", "eq"))))),
    (
        "?",
        (
            ("classes", ("capital",)),
            ("attributes", (("latex", "Cref"), ("rst", "numref"))),
        ),
    ),
    ("&", (("classes", ()), ("attributes", (("latex", "gls"), ("rst", "gls"))))),
    (
        "%",
        (
            ("classes", ("capital",)),
            ("attributes", (("latex", "Gls"), ("rst", "glsc"))),
        ),
    ),
)

PREFIX_MAP_LATEX_R = (
    ("cref", "+"),
    ("Cref", "?"),
    ("ref", "!"),
    ("eqref", "="),
    ("cite", ""),
    ("gls", "&"),
    ("Gls", "%"),
)
PREFIX_MAP_RST_R = (
    ("numref", "+"),
    ("ref", "!"),
    ("eq", "="),
    ("cite", ""),
    ("gls", "&"),
    ("glsc", "%"),
)

CITE_HTML_NAMES = (("Math", "eqn."), ("Image", "fig."), ("Table", "tbl."))

RST_KNOWN_ROLES = (
    "py:attr",
    "py:meth",
    "py:class",
    "py:func",
    "py:mod",
    "attr",
    "meth",
    "class",
    "func",
    "mod",
    "download",
    "doc",
    "file",
    "program",
)



from copy import deepcopy
import six

try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser


class HTML2JSONParser(HTMLParser, object):
    """parses html content to a JSON object,
    of the form::

        {"attrs": {}, "data": [], "children": []}

    """

    _tag_key = "1_tag"
    _data_key = "2_data"
    _tag_attrs_key = "3_attributes"
    _children_key = "4_children"

    def __init__(
        self,
        ignore_tags=("head", "script", "style"),
        ignore_classes=("footer", "sphinxsidebar", "clearer"),
        rstrip_data=True,
        sort_class_attr=True,
        replace_data_lines=None,
        convert_charrefs=False,
    ):
        """parses html content to a JSON object,
        of the form::

            {"tag": "tag_name", "attrs": {}, "data": [], "children": []}

        Parameters
        ----------
        ignore_tags : list[str]
            HTML tags that will be ignored (and all their children)
        ignore_classes : list[str]
            HTML tags with one or more of these classes will be ignored (and all their children)
        rstrip_data : bool
            apply `rstrip()` to data text, and don't add if data == ''
        sort_class_attr : bool
            if an attribute is named 'class', its contents will be split and sorted
        replace_data_lines: None or dict
            mapping of data lines to replace (useful for mapping across different environment versions)
        convert_charrefs: bool
            If True, all character references (except the ones in script/style elements)
            are automatically converted to the corresponding Unicode characters.

        """
        if six.PY2:
            super(HTML2JSONParser, self).__init__()
        else:
            super(HTML2JSONParser, self).__init__(convert_charrefs=convert_charrefs)
        self._content = {}
        self._curr_tags = []
        self._curr_depth = 0
        self._ignore_depth = None
        self._rstrip_data = rstrip_data
        self._sort_class_attr = sort_class_attr
        self._ignore_tags = ignore_tags
        self._ignore_classes = set(ignore_classes)
        self._replace_data_lines = replace_data_lines or {}

    @property
    def parsed(self):
        return deepcopy(self._content)

    def reset(self):
        self._content = {}
        self._curr_tags = []
        self._curr_depth = 0
        self._ignore_depth = None
        super(HTML2JSONParser, self).reset()

    def _get_subcontent(self):
        sub_content = self._content
        for ptag in self._curr_tags:
            sub_content = sub_content[self._children_key][ptag]
        return sub_content

    def handle_starttag(self, tag, attrs):
        self._curr_depth += 1
        attr_dict = dict(attrs)
        classes = attr_dict.get("class", "").split()
        if self._ignore_depth is not None:
            return
        elif tag in self._ignore_tags or self._ignore_classes.intersection(classes):
            # we ignore any tags and data, until the current depth is less than the ignore depth
            self._ignore_depth = self._curr_depth
            return
        sub_content = self._get_subcontent()
        if self._sort_class_attr and "class" in attr_dict:
            attr_dict["class"] = sorted(classes)
        tag_dict = {self._tag_key: tag}
        if attr_dict:
            tag_dict[self._tag_attrs_key] = attr_dict
        sub_content.setdefault(self._children_key, []).append(tag_dict)
        self._curr_tags.append(len(sub_content[self._children_key]) - 1)

    def handle_endtag(self, tag):
        self._curr_depth -= 1

        if self._ignore_depth is not None:
            if self._ignore_depth > self._curr_depth:
                # print("exited ignore: {}".format(tag))
                self._ignore_depth = None
            return

        if tag != self._get_subcontent()[self._tag_key]:
            raise AssertionError(
                "{} != {} (current path: {})".format(
                    tag, self._get_subcontent()[self._tag_key], self._curr_tags
                )
            )
        self._curr_tags = self._curr_tags[:-1]

    def handle_data(self, data):
        if self._ignore_depth is not None:
            return
        if not data.strip():
            return
        data = data.splitlines()
        if self._rstrip_data:
            data = [d.rstrip() for d in data]
        if self._replace_data_lines:
            data = [self._replace_data_lines.get(d, d) for d in data]
        sub_content = self._get_subcontent()
        sub_content.setdefault(self._data_key, []).extend(data)

    def handle_entityref(self, name):
        pass

    def handle_charref(self, name):
        pass

    def handle_comment(self, data):
        pass



import os


def get_test_source_dir(subfolder=None):
    if subfolder is None:
        return os.path.abspath(os.path.join(os.path.dirname(__file__), "sourcedirs"))
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), "sourcedirs", subfolder)
    )



"""pandoc filters used in converting markdown to a target format

Other sources of information:

- [Pandoc User Guide](https://pandoc.org/MANUAL.html#citations)
- [List of Pandoc Elements](https://metacpan.org/pod/Pandoc::Elements)
- [reStructuredText Primer](http://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html)
- [reStructuredText Directives](http://docutils.sourceforge.net/docs/ref/rst/directives.html#figure)
- [sphinxcontrib-bibtex](https://sphinxcontrib-bibtex.readthedocs.io/en/latest/usage.html)

"""  # noqa: E501
import panflute as pf

from ipypublish.filters_pandoc.definitions import IPUB_META_ROUTE
from ipypublish.filters_pandoc.utils import apply_filter, get_option, create_ipub_meta
from ipypublish.filters_pandoc import (
    prepare_cites,
    prepare_labels,
    prepare_raw,
    format_cite_elements,
    format_label_elements,
    format_raw_spans,
    rmarkdown_to_mpe,
)


def pandoc_filters():
    """ run a set of ipypublish pandoc filters directly on the pandoc AST,
    via ``pandoc --filter ipubpandoc``
    """
    doc = pf.load()

    # in an rmarkdown file, the metadata will be under a root `jupyter` key
    jmeta = doc.get_metadata("jupyter", {})
    meta = pf.tools.meta2builtin(doc.metadata)
    if "jupyter" in meta and hasattr(meta["jupyter"], "items"):
        jmeta = meta.pop("jupyter")
        meta.update(jmeta)
        doc.metadata = meta  # builtin2meta(meta)

    apply_filters = doc.get_metadata(IPUB_META_ROUTE + ".apply_filters", default=True)
    convert_raw = doc.get_metadata(IPUB_META_ROUTE + ".convert_raw", default=True)

    if apply_filters:
        if convert_raw:
            filters = [
                prepare_raw.main,
                prepare_cites.main,
                prepare_labels.main,
                format_cite_elements.main,
                format_raw_spans.main,
                format_label_elements.main,
                rmarkdown_to_mpe.main,
            ]
        else:
            filters = [
                prepare_cites.main,
                prepare_labels.main,
                format_cite_elements.main,
                format_label_elements.main,
                rmarkdown_to_mpe.main,
            ]
    else:
        filters = []

    out_doc = doc
    for func in filters:
        out_doc = func(out_doc)  # type: pf.Doc
    # TODO strip meta?
    pf.dump(doc)


def jinja_filter(
    source, to_format, nb_metadata, cell_metadata, from_format="markdown", strip=True
):
    """run a set of ipypublish pandoc filters as a Jinja2 filter

    We convert the source to an intermediary pandoc-json AST format,
    run the pandocfilters, then convert to the to_format

    Parameters
    ----------
    source: str
        content to convert
    to_format: str
        format of output
    nb_metadata: dict
        mapping of notebook level metadata
    cell_metadata: dict
        mapping of cell level metadata
    from_format: str
        format of source
    strip: bool
        strip any blank lines from the start/end of the final string

    Notes
    -----

    **Available Meta Options**

    The following options are available in {"ipub": {}}

    apply_filters=True: bool
        apply filters to markdown
    convert_raw=True: bool
        if True attempt to extract non-markdown formats
        and convert them to the target format, e.g. rst roles to latex tags
    at_notation=True: bool
        interpret @label as a reference type based on its prefix modifier,
        latex: '' = cite '+' = cref,    '^' = Cref,    '!' = ref,  '=' = eqref
        rst: '' = :cite: '+' = :numref: '^' = :numref: '!' = :ref: '=' = :eq:
    reftag="cite": str
        default latex tag for references
    use_numref=True: bool
        whether to use the ``:numref:`` role or just ``:ref:``
        ``:numref:`` requires ``numfig = True`` in conf.py and,
        for section numbering, a toc tree with ``:numbered:``
    strip_meta=True: bool
        if True strip any source metadata, contained in the top matter

    The options will be taken in order of preference from:
    source.metadata > cell.metadata > nb.metadata

    For source.metadata, see https://pandoc.org/MANUAL.html#metadata-blocks:

    .. code-block:: yaml

        ---
        ipub:
          pandoc:
            use_numref: True
        ---
        +@label

    """
    if not source.strip():
        return source.strip() if strip else source

    # convert the source to a format agnostic Doc
    doc = apply_filter(source, dry_run=True)  # type: pf.Doc

    # find the preferential versions of the metadata values
    # TODO a make this autopopulate (possibly from schema)
    option_preference = [doc.metadata, cell_metadata, nb_metadata]
    apply_filters = get_option(
        option_preference, keypath=IPUB_META_ROUTE + ".apply_filters", default=True
    )
    convert_raw = get_option(
        option_preference, keypath=IPUB_META_ROUTE + ".convert_raw", default=True
    )
    hide_raw = get_option(
        option_preference, keypath=IPUB_META_ROUTE + ".hide_raw", default=False
    )
    numref = get_option(
        option_preference, keypath=IPUB_META_ROUTE + ".use_numref", default=True
    )
    at_notation = get_option(
        option_preference, keypath=IPUB_META_ROUTE + ".at_notation", default=True
    )
    reftag = get_option(
        option_preference, keypath=IPUB_META_ROUTE + ".reftag", default="cite"
    )
    strip_meta = get_option(
        option_preference, keypath=IPUB_META_ROUTE + ".strip_meta", default=True
    )

    if apply_filters:
        # TODO store the original metadata and replace it at end?
        # original_meta = copy.deepcopy(doc.metadata)

        # set metadata with preferential values
        meta = pf.tools.meta2builtin(doc.metadata)
        meta.update(
            create_ipub_meta(
                {
                    "use_numref": numref,
                    "at_notation": at_notation,
                    "reftag": reftag,
                    "hide_raw": hide_raw,
                }
            )
        )
        doc.metadata = meta  # builtin2meta(meta)

        # doc.metadata["ipub"]["use_numref"] = builtin2meta(numref)
        # doc.metadata["ipub"]["at_notation"] = builtin2meta(at_notation)
        # doc.metadata["ipub"]["reftag"] = builtin2meta(reftag)

        # set filters
        if convert_raw:
            filters = [
                prepare_raw.main,
                prepare_cites.main,
                prepare_labels.main,
                format_cite_elements.main,
                format_raw_spans.main,
                format_label_elements.main,
            ]
        else:
            filters = [
                prepare_cites.main,
                prepare_labels.main,
                format_cite_elements.main,
                format_label_elements.main,
            ]
    else:
        filters = []

    out_str = apply_filter(
        doc,
        filters,
        in_format=from_format,
        out_format=to_format,
        strip_meta=bool(strip_meta),
    )
    if strip:
        out_str = out_str.strip()

    return out_str



import logging
import os

import traitlets as traits
from nbconvert.preprocessors import Preprocessor
from nbformat.notebooknode import NotebookNode


class LatexDocHTML(Preprocessor):
    r""" processing of ipub metatags, specific to html

    - import embedded html files
    - add refmap key to references for {label:reference name} lookup
      e.g. {"fig:test":"fig. 1"}
    - add caption_prefix tag for floats with correct numbering/name
      e.g. cell.metadata.ipub.figure.caption_prefix = "Figure 1: "

    """

    metapath = traits.Unicode("", help="the path to the meta data").tag(config=True)
    filesfolder = traits.Unicode("", help="the folder to point towards").tag(
        config=True
    )
    src_name = traits.Unicode(
        "src",
        help=("for embedding, if reveal js slides use data-src " "(for lazy loading)"),
    ).tag(config=True)

    @traits.validate("src_name")
    def _valid_value(self, proposal):
        if proposal["value"] not in ["src", "data-src"]:
            raise traits.TraitError("src_name must be src or src or data-src")
        return proposal["value"]

    def __init__(self, *args, **kwargs):
        super(LatexDocHTML, self).__init__(*args, **kwargs)

    def resolve_path(self, fpath, filepath):
        """resolve a relative path, w.r.t. another filepath """
        if not os.path.isabs(fpath):
            fpath = os.path.join(os.path.dirname(str(filepath)), fpath)
            fpath = os.path.abspath(fpath)
        return fpath

    def embed_html(self, cell, path):
        """ a new cell, based on embedded html file
        """
        logging.info("embedding html in notebook from: {}".format(path))

        height = int(cell.metadata.ipub.embed_html.get("height", 0.5) * 100)
        width = int(cell.metadata.ipub.embed_html.get("width", 0.5) * 100)
        embed_code = (
            '<iframe style="display:block; margin: 0 auto; height:{height}vh; '
            'width:{width}vw; overflow:auto; resize:both" {src}="{path}" '
            'frameborder="0" allowfullscreen>'
            "</iframe>"
        ).format(src=self.src_name, path=path, height=height, width=width)

        # add to the exising output or create a new one
        if cell.outputs:
            cell.outputs[0]["data"]["text/html"] = embed_code
        else:
            cell.outputs.append(
                NotebookNode(
                    {
                        "data": {"text/html": embed_code},
                        "execution_count": 0,
                        "metadata": {},
                        "output_type": "execute_result",
                    }
                )
            )

        return cell

    def preprocess(self, nb, resources):

        logging.info(
            "processing notebook for html output"
            + " in ipub metadata to: {}".format(self.metapath)
        )

        final_cells = []
        float_count = dict(
            [("figure", 0), ("table", 0), ("code", 0), ("text", 0), ("error", 0)]
        )
        for i, cell in enumerate(nb.cells):
            if hasattr(cell.metadata, "ipub"):
                if hasattr(cell.metadata.ipub, "embed_html"):
                    if hasattr(cell.metadata.ipub.embed_html, "filepath"):
                        paths = [cell.metadata.ipub.embed_html.filepath]
                        if hasattr(cell.metadata.ipub.embed_html, "other_files"):
                            if not isinstance(
                                cell.metadata.ipub.embed_html.other_files, list
                            ):
                                raise TypeError(
                                    "{} should be a list".format(
                                        cell.metadata.ipub.embed_html.other_files
                                    )
                                )
                            paths += cell.metadata.ipub.embed_html.other_files
                        for j, path in enumerate(paths):
                            fpath = self.resolve_path(path, self.metapath)
                            if not os.path.exists(fpath):
                                logging.warning(
                                    "file in embed html metadata doesn't exist"
                                    ": {}".format(fpath)
                                )
                            else:
                                resources.setdefault("external_file_paths", [])
                                resources["external_file_paths"].append(fpath)
                                if j == 0:
                                    self.embed_html(
                                        cell,
                                        os.path.join(
                                            self.filesfolder, os.path.basename(fpath)
                                        ),
                                    )

                    elif hasattr(cell.metadata.ipub.embed_html, "url"):
                        self.embed_html(cell, cell.metadata.ipub.embed_html.url)
                    else:
                        logging.warning(
                            "cell {} has no filepath or url key in its "
                            "metadata.embed_html".format(i)
                        )

                for floattype, floatabbr in [
                    ("figure", "fig."),
                    ("table", "tbl."),
                    ("code", "code"),
                    ("text", "text"),
                    ("error", "error"),
                ]:
                    if floattype in cell.metadata.ipub:
                        if floattype != "code" and not cell.get("outputs", []):
                            continue
                        float_count[floattype] += 1
                        if not isinstance(cell.metadata.ipub[floattype], dict):
                            continue
                        cell.metadata.ipub[floattype][
                            "caption_prefix"
                        ] = "<b>{0} {1}:</b> ".format(  # noqa: E501
                            floattype.capitalize(), float_count[floattype]
                        )
                        if "label" in cell.metadata.ipub[floattype]:
                            label = "{0} {1}".format(floatabbr, float_count[floattype])
                            resources.setdefault("refmap", {})[
                                cell.metadata.ipub[floattype]["label"]
                            ] = label

            final_cells.append(cell)
        nb.cells = final_cells

        return nb, resources



""" a panflute filter to find citations in markdown,
with attributes and prefixes,

- extract the attributes, classes and prefix
- wrap the cite with a Span with class 'attribute-Cite',
- add the classes, attributes and prefix to the Span, and
- remove the attribute string and prefix

For example:

```
+@label {}.class-name a=1} xyz *@label2* @label3{ .b}
```

would be converted to this html:

```html
<p>
<span class="class-name attribute-Cite" data-a="1" data-prefix="+">
<span class="citation" data-cites="label">@label</span>
</span>
xyz
<em><span class="citation" data-cites="label2">@label2</span></em>
<span class="b attribute-Cite" data-prefix="{">
<span class="citation" data-cites="label3">@label3</span>
</span>
</p>
```

Optionally, this can be turned off by adding to the Document metadata

meta["ipub"]["pandoc"]["at_notation"] = False

"""
from panflute import Element, Doc, Cite, RawInline, Link  # noqa: F401
import panflute as pf


from ipypublish.filters_pandoc.utils import find_attributes, get_pf_content_attr
from ipypublish.filters_pandoc.definitions import (
    ATTRIBUTE_CITE_CLASS,
    IPUB_META_ROUTE,
    PREFIX_MAP,
)


def process_citations(element, doc):
    # type: (Element, Doc) -> Element
    if not doc.get_metadata(IPUB_META_ROUTE + ".at_notation", True):
        return None

    content_attr = get_pf_content_attr(element, pf.Cite)
    if not content_attr:
        return None
    initial_content = getattr(element, content_attr)

    if not initial_content:
        return None

    final_content = []
    skip = 0
    for subel in initial_content:

        if skip:
            skip -= 1
            continue

        if not isinstance(subel, pf.Cite):
            final_content.append(subel)
            continue

        classes = []
        attributes = {}
        append = None

        # check if the cite has a valid prefix, if so extract it
        if (
            isinstance(subel.prev, pf.Str)
            and subel.prev.text
            and (subel.prev.text[-1] in dict(PREFIX_MAP))
        ):

            prefix = subel.prev.text[-1]
            mapping = dict(dict(PREFIX_MAP)[prefix])
            classes.extend(mapping["classes"])
            attributes.update(mapping["attributes"])

            # remove prefix from preceding string
            string = final_content.pop()
            if len(string.text) > 1:
                final_content.append(pf.Str(string.text[:-1]))

        # check if the cite has a preceding class/attribute container
        attr_dict = find_attributes(subel, allow_space=True)
        if attr_dict:
            classes.extend(attr_dict["classes"])
            attributes.update(attr_dict["attributes"])
            skip = len(attr_dict["elements"])
            append = attr_dict["append"]

        if classes or attributes:
            classes.append(ATTRIBUTE_CITE_CLASS)
            final_content.append(
                pf.Span(subel, classes=sorted(set(classes)), attributes=attributes)
            )
        else:
            final_content.append(subel)

        if append:
            final_content.append(append)

    setattr(element, content_attr, final_content)
    return element


def prepare(doc):
    # type: (Doc) -> None
    pass


def finalize(doc):
    # type: (Doc) -> None
    pass


def main(doc=None, extract_formats=True):
    # type: (Doc) -> None
    """if extract_formats then convert citations defined in
    latex, rst or html formats to special Span elements
    """
    return pf.run_filter(process_citations, prepare, finalize, doc=doc)


if __name__ == "__main__":
    main()



""" a panflute filter to format Cite elements

The :py:mod:`ipypublish.filters_pandoc.prepare_cites` filter should be run
first to access the functionality below:

"""
from panflute import Element, Doc, Span, Cite  # noqa: F401
import panflute as pf

from ipypublish.filters_pandoc.definitions import (
    ATTRIBUTE_CITE_CLASS,
    CONVERTED_CITE_CLASS,
    IPUB_META_ROUTE,
    CITE_HTML_NAMES,
)
from ipypublish.filters_pandoc.html_bib import read_bibliography, process_bib_entry


def format_cites(cite, doc):
    # type: (Cite, Doc) -> Element
    """
    originally adapted from:
    `pandoc-xnos <https://github.com/tomduck/pandoc-xnos/>`_
    """
    if not isinstance(cite, pf.Cite):
        return None

    # default tags for latex and rst
    cite_tag = doc.get_metadata(IPUB_META_ROUTE + ".reftag", "cite")
    cite_role = "cite"
    html_capitalize = False

    # check is the Cite has a surrounding Span to supply attributed
    span = None
    if isinstance(cite.parent, pf.Span) and ATTRIBUTE_CITE_CLASS in cite.parent.classes:
        span = cite.parent

        cite_tag = span.attributes.get("latex", cite_tag)
        cite_role = span.attributes.get("rst", cite_role)
        html_capitalize = "capital" in span.classes

    if cite_role == "numref" and (
        not doc.get_metadata(IPUB_META_ROUTE + ".use_numref", False)
    ):
        cite_role = "ref"

    if doc.format in ("latex", "tex"):

        if cite_tag in ["cite", "cref", "Cref"] or len(cite.citations) == 1:
            # multiple labels are allowed
            return pf.RawInline(
                "\\{0}{{{1}}}".format(
                    cite_tag, ",".join([c.id for c in cite.citations])
                ),
                format="tex",
            )
        else:
            tex = ", ".join(
                ["\\{0}{{{1}}}".format(cite_tag, c.id) for c in cite.citations[:-1]]
            ) + " and \\{0}{{{1}}}".format(cite_tag, cite.citations[-1].id)
            return pf.RawInline(tex, format="tex")

    if doc.format == "rst":
        if len(cite.citations) == 1:
            raw = pf.RawInline(
                ":{0}:`{1}`".format(cite_role, cite.citations[0].id), format="rst"
            )
        elif cite_role == "cite":
            raw = pf.RawInline(
                ":{0}:`{1}`".format(
                    cite_role, ",".join([c.id for c in cite.citations])
                ),
                format="rst",
            )
        else:
            raw = pf.RawInline(
                ", ".join(
                    [":{0}:`{1}`".format(cite_role, c.id) for c in cite.citations[:-1]]
                )
                + " and :{0}:`{1}`".format(cite_role, cite.citations[-1].id),
                format="rst",
            )

        # in testing, rst cite roles required space either side
        # to render correctly
        # TODO check if spacing is required for :cite: roles (and others)
        if cite_role == "cite":
            elem = span if span else cite  # type: pf.Inline
            raw = [raw]
            if elem.prev and not isinstance(elem.prev, pf.Space):
                raw.insert(0, pf.Space())
            if elem.next and not isinstance(elem.next, pf.Space):
                raw.append(pf.Space())

        return raw

    if doc.format in ("html", "html5"):

        elements = []
        cites = set()
        names = dict()
        unknown = set()

        for citation in cite.citations:
            ref = doc.get_metadata("$$references.{}".format(citation.id), False)
            if ref:
                # ref -> e.g. {"type": "Math", "number": 1}
                prefix = dict(CITE_HTML_NAMES).get(ref["type"], ref["type"])
                prefix = prefix.capitalize() if html_capitalize else prefix

                # label = "{} {}".format(prefix, ref["number"])
                # elements.append(pf.RawInline(
                #     '<a href="#{0}">{1}</a>'.format(citation.id, label),
                #     format=doc.format))
                # found_ref = True

                names.setdefault(prefix, set()).add(
                    '<a href="#{0}">{1}</a>'.format(citation.id, ref["number"])
                )

            elif citation.id in doc.bibdatabase:
                cites.add(process_bib_entry(citation.id, doc.bibdatabase, doc.bibnums))
                # elements.append(pf.RawInline(
                #     process_bib_entry(
                #         citation.id, doc.bibdatabase, doc.bibnums),
                #     format=doc.format))
                # found_ref = True
            else:
                unknown.add(citation.id)
                # elements.append(pf.Cite(citations=[citation]))

        # if found_ref:
        #     return elements
        # else:
        # return pf.RawInline(
        #     '<span style="background-color:rgba(225, 0, 0, .5)">'
        #     # 'No reference found for: {}</span>'.format(
        #     '{}</span>'.format(
        #         ", ".join([c.id for c in cite.citations])))

        elements = []
        if cites:
            # TODO sort
            elements.append(
                pf.RawInline(
                    "<span>[{}]</span>".format(",".join(c for c in cites)),
                    format=doc.format,
                )
            )
        if names:
            # TODO sort
            for prefix, labels in names.items():
                elements.append(
                    pf.RawInline(
                        "<span>{} {}</span>".format(
                            prefix, ",".join(l for l in labels)
                        ),
                        format=doc.format,
                    )
                )
        if unknown:
            elements.append(
                pf.RawInline(
                    '<span style="background-color:rgba(225, 0, 0, .5)">'
                    # 'No reference found for: {}</span>'.format(
                    "{}</span>".format(", ".join([l for l in unknown]))
                )
            )

        return elements


def format_span_cites(span, doc):
    # type: (Cite, Doc) -> Element
    if not isinstance(span, pf.Span):
        return None

    if CONVERTED_CITE_CLASS not in span.classes:
        return None

    if doc.format in ("latex", "tex"):
        cite_tag = "cite"
        if span.attributes["format"] == "latex":
            cite_tag = span.attributes["tag"]
        # TODO use cref for rst ref/numref
        return pf.RawInline(
            "\\{0}{{{1}}}".format(cite_tag, span.identifier, format="tex")
        )

    if doc.format == "rst":
        cite_role = "cite"
        if span.attributes["format"] == "rst":
            cite_role = span.attributes["role"]
        # TODO use ref for latex ref/cref/Cref
        return [
            pf.RawInline(":{0}:`{1}`".format(cite_role, span.identifier), format="rst")
        ]

    if doc.format in ("html", "html5"):
        # <cite data-cite="cite_key">text</cite>
        return (
            [
                pf.RawInline(
                    '<cite data-cite="{}">'.format(span.identifier), format="html"
                )
            ]
            + list(span.content)
            + [pf.RawInline("</cite>", format="html")]
        )


def prepare(doc):
    # type: (Doc) -> None
    doc.bibnums = {}
    doc.bibdatabase = {}
    if doc.format in ("html", "html5"):
        bib_path = doc.get_metadata("ipub.bibliography", None)
        if bib_path:
            doc.bibdatabase = read_bibliography(bib_path)


def finalize(doc):
    # type: (Doc) -> None
    del doc.bibnums
    del doc.bibdatabase


def strip_cite_spans(span, doc):
    # type: (Span, Doc) -> Element
    if isinstance(span, pf.Span) and ATTRIBUTE_CITE_CLASS in span.classes:
        return list(span.content)


def main(doc=None, strip_spans=True):
    # type: (Doc) -> None
    to_run = [format_cites]
    if strip_spans:
        to_run.append(strip_cite_spans)
    return pf.run_filters(to_run, prepare, finalize, doc=doc)


if __name__ == "__main__":
    main()



import pytest
from ipypublish.convert import nbmerge


@pytest.mark.ipynb("basic_nb")
def test_nbmerge_one_notebook(ipynb_app):
    nb, path = nbmerge.merge_notebooks(ipynb_app.source_path)
    assert nb.metadata.test_name == "notebook1"
    assert len(nb.cells) == 2


@pytest.mark.ipynb("merge_nbs", main_file=None)
def test_nbmerge_two_notebooks(ipynb_app):
    nb, path = nbmerge.merge_notebooks(ipynb_app.source_path)
    assert nb.metadata.test_name == "notebook1"
    assert len(nb.cells) == 4


@pytest.mark.ipynb("merge_nbs", main_file=None)
def test_merge_latex(ipynb_app):
    ipynb_app.run()
    ipynb_app.assert_converted_exists()



import os
import json
import inspect
import importlib
import re
import pkg_resources

from six import string_types
import ruamel.yaml as yaml

# python 2/3 compatibility
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib


def handle_error(msg, err_type, logger, raise_msg=None, log_msg=None):
    """handle an error, by logging it, then raising an exception"""
    if raise_msg is None:
        raise_msg = msg
    if log_msg is None:
        log_msg = msg

    logger.error(log_msg)
    raise err_type(raise_msg)


def read_file_from_directory(
    dir_path,
    file_name,
    jtype,
    logger,
    interp_ext=False,
    ext_types=(("json", (".json",)), ("yaml", (".yaml", ".yaml.j2"))),
):
    """load a file situated in a directory

    if ``interp_ext=True``:
    interpret the file extension *via* ``ext_types``
    and load file in a suitable manner

    """
    if isinstance(dir_path, string_types):
        dir_path = pathlib.Path(dir_path)

    file_path = dir_path.joinpath(file_name)

    if not file_path.exists():
        handle_error(
            "the {} does not exist: {}".format(jtype, file_path), IOError, logger=logger
        )

    ext_type = None
    ext_map = {ext: ftype for ftype, exts in ext_types for ext in exts}
    # Place longer extensions first to keep shorter ones from matching first
    for ext in sorted(ext_map.keys(), key=len, reverse=True):
        if file_name.endswith(ext):
            ext_type = ext_map[ext]

    if ext_type is not None and interp_ext:
        with file_path.open() as fobj:
            try:
                if ext_type == "json":
                    data = json.load(fobj)
                elif ext_type == "yaml":
                    data = yaml.safe_load(fobj)
                else:
                    raise ValueError("extension type not recognised")
            except Exception as err:
                handle_error(
                    "failed to read {} ({}): {}".format(jtype, file_path, err),
                    IOError,
                    logger=logger,
                )
    else:
        with file_path.open() as fobj:
            data = fobj.read()

    return data


def get_module_path(module):
    """return a directory path to a module"""
    return pathlib.Path(os.path.dirname(os.path.abspath(inspect.getfile(module))))


def read_file_from_module(
    module_path,
    file_name,
    jtype,
    logger,
    interp_ext=False,
    ext_types=(("json", (".json")), ("yaml", (".yaml", ".yaml.j2", "yaml.tex.j2"))),
):
    """load a file situated in a python module

    if ``interp_ext=True``:
    interpret the file extension *via* ``ext_type``
    and load file in a suitable manner

    """
    try:
        outline_module = importlib.import_module(module_path)
    except ModuleNotFoundError:  # noqa: F821
        handle_error(
            "module {} containing {} {} not found".format(
                module_path, jtype, file_name
            ),
            ModuleNotFoundError,
            logger=logger,
        )  # noqa: F821

    return read_file_from_directory(
        get_module_path(outline_module),
        file_name,
        jtype,
        logger,
        interp_ext=interp_ext,
        ext_types=ext_types,
    )


def get_valid_filename(s):
    """
    Return the given string converted to a string that can be used for a clean
    filename. Remove leading and trailing spaces; convert other spaces to
    underscores; and remove anything that is not an alphanumeric, dash,
    underscore, or dot.
    >>> get_valid_filename("john's portrait in 2004.jpg")
    'johns_portrait_in_2004.jpg'
    """
    s = str(s).strip().replace(" ", "_")
    return re.sub(r"(?u)[^-\w.]", "", s)


def find_entry_point(name, group, logger, preferred=None):
    """find an entry point by name and group

    Parameters
    ----------
    name: str
        name of entry point
    group: str
        group of entry point
    preferred: str
        if multiple matches are found, prefer one from this module

    """
    entry_points = list(pkg_resources.iter_entry_points(group, name))
    if len(entry_points) == 0:
        handle_error(
            "The {0} entry point " "{1} could not be found".format(group, name),
            pkg_resources.ResolutionError,
            logger,
        )
    elif len(entry_points) != 1:
        # default to the preferred package
        oentry_points = []
        if preferred:
            oentry_points = [
                ep for ep in entry_points if ep.module_name.startswith(preferred)
            ]
        if len(oentry_points) != 1:
            handle_error(
                "Multiple {0} plugins found for "
                "{1}: {2}".format(group, name, entry_points),
                pkg_resources.ResolutionError,
                logger,
            )
        logger.info(
            "Multiple {0} plugins found for {1}, "
            "defaulting to the {2} version".format(group, name, preferred)
        )
        entry_point = oentry_points[0]
    else:
        entry_point = entry_points[0]
    return entry_point.load()



import copy
import io
import logging
import os

import bibtexparser

from ipypublish.bib2glossary.definitions import (
    ETYPE_GLOSS,
    ETYPE_ACRONYM,
    ETYPE_SYMBOL,
    NEWGLOSS_FIELDS,
    NEWACRONYM_FIELDS,
)

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

logger = logging.getLogger(__name__)


class BibGlossEntry(object):
    _allowed_types = (ETYPE_GLOSS, ETYPE_ACRONYM, ETYPE_SYMBOL)

    def __init__(self, entry_dict):

        self._validate_dict(entry_dict)
        self._entry_dict = entry_dict

    def _validate_dict(self, dct):
        if "ID" not in dct:
            raise KeyError
        if "ENTRYTYPE" not in dct:
            raise KeyError

        if dct["ENTRYTYPE"] not in self._allowed_types:
            raise TypeError("ENTRYTYPE must be one of: {}".format(self._allowed_types))

        if dct["ENTRYTYPE"] == ETYPE_ACRONYM:
            if "abbreviation" not in dct or "longname" not in dct:
                raise KeyError
        elif dct["ENTRYTYPE"] == ETYPE_GLOSS or dct["ENTRYTYPE"] == ETYPE_SYMBOL:
            if "name" not in dct or "description" not in dct:
                raise KeyError

    def _get_key(self):
        return self._entry_dict["ID"]

    def _set_key(self, key):
        self._entry_dict["ID"] = key

    key = property(_get_key, _set_key)

    @property
    def type(self):
        return self._entry_dict["ENTRYTYPE"]

    def __contains__(self, key):
        return key in self._entry_dict

    def get(self, key):
        return self._entry_dict[key]

    @property
    def label(self):
        if self.type == ETYPE_ACRONYM:
            return self.get("abbreviation")
        elif self.type == ETYPE_GLOSS:
            return self.get("name")
        elif self.type == ETYPE_SYMBOL:
            return self.get("name")
        else:
            raise NotImplementedError

    @property
    def sortkey(self):
        if "sort" in self:
            return self.get("sort")
        else:
            return self.label.lower()

    @property
    def plural(self):
        if "plural" in self:
            return self.get("plural")
        else:
            return "{}s".format(self.label)

    @property
    def text(self):
        if self.type == ETYPE_ACRONYM:
            return self.get("longname")
        elif self.type == ETYPE_GLOSS:
            return self.get("description")
        elif self.type == ETYPE_SYMBOL:
            return self.get("description")
        else:
            raise NotImplementedError

    def __repr__(self):
        return "BibGlossEntry(key={0},label={1})".format(self.key, self.label)

    def to_dict(self):
        return copy.deepcopy(self._entry_dict)

    def to_latex(self):

        if self.type in [ETYPE_GLOSS, ETYPE_SYMBOL]:
            options = []
            for field in sorted(NEWGLOSS_FIELDS):
                if field in self:
                    options.append("{0}={{{1}}}".format(field, self.get(field)))
            if self.type == ETYPE_SYMBOL:
                options.append("type={symbols}")
            body = "{{{key}}}{{\n    {options}\n}}".format(
                key=self.key, options=",\n    ".join(options)
            )
            return "\\newglossaryentry" + body

        elif self.type == ETYPE_ACRONYM:
            body = "{{{key}}}{{{abbrev}}}{{{long}}}".format(
                key=self.key, abbrev=self.label, long=self.text
            )
            options = []
            for field in sorted(NEWACRONYM_FIELDS):
                if field in self:
                    options.append("{0}={{{1}}}".format(field, self.get(field)))
            if options:
                body = "[" + ",".join(options) + "]" + body

            return "\\newacronym" + body


class BibGlossDB(MutableMapping):
    def __init__(self):
        self._entries = {}

    def __getitem__(self, key):
        return self._entries[key]

    def __setitem__(self, key, entry):
        if not isinstance(entry, BibGlossEntry):
            raise ValueError("value must be a BibGlossEntry")
        if key != entry.key:
            raise ValueError("key must equal entry.key")
        self._entries[key] = entry

    def __delitem__(self, key):
        del self._entries[key]

    def __iter__(self):
        return iter(self._entries)

    def __len__(self):
        return len(self._entries)

    @staticmethod
    def get_fake_entry_obj(key):
        return BibGlossEntry(
            {"ENTRYTYPE": ETYPE_GLOSS, "ID": key, "name": key, "description": ""}
        )

    def load_bib(
        self,
        text_str=None,
        path=None,
        bibdb=None,
        encoding="utf8",
        ignore_nongloss_types=False,
        ignore_duplicates=False,
    ):
        """load a bib file

        Parameters
        ----------
        text_str=None: str or None
            string representing the bib file contents
        path=None: str or None
            path to bibfile
        bibdb=None: bibtexparser.bibdatabase.BibDatabase or None
        encoding="utf8": str
            bib file encoding
        ignore_nongloss_types: bool
            if False, a KeyError will be raised for non-gloss types
        ignore_duplicates: bool
            if False, a KeyError will be raised if multiple entries are found
            with the same key, otherwise only the first entry will be used

        """
        bib = None

        if sum([e is not None for e in [text_str, path, bibdb]]) != 1:
            raise ValueError("only one of text_str, path or bib must be supplied")
        if bibdb is not None:
            if not isinstance(bibdb, bibtexparser.bibdatabase.BibDatabase):
                raise ValueError("bib is not a BibDatabase instance")
            bib = bibdb
        elif path is not None:
            if text_str is not None:
                raise ValueError("text_str and path cannot be set at the same time")
            with io.open(path, encoding=encoding) as fobj:
                text_str = fobj.read()

        if bib is None:
            parser = bibtexparser.bparser.BibTexParser()
            parser.ignore_nonstandard_types = False
            parser.encoding = encoding
            bib = parser.parse(text_str)
            # TODO doesn't appear to check for key duplication
            # see https://github.com/sciunto-org/python-bibtexparser/issues/237

        entries = {}
        for entry_dict in bib.entries:

            try:
                entry = BibGlossEntry(entry_dict)
            except TypeError:
                if ignore_nongloss_types:
                    logger.warning("Skipping non-glossary entry")
                    continue
                else:
                    raise

            if entry.key in entries:
                if ignore_duplicates:
                    logger.warning("Skipping duplicate key entry")
                    continue
                else:
                    raise KeyError(
                        "the bib file contains "
                        "multiple entries with the key: {}".format(entry.key)
                    )

            entries[entry.key] = entry

        # self._bib = bib
        self._entries = entries

        return True

    def load_tex(
        self,
        text_str=None,
        path=None,
        encoding="utf8",
        skip_ioerrors=False,
        ignore_unknown_types=True,
    ):
        """load a tex file

        Parameters
        ----------
        text_str=None: str or None
            string representing the bib file contents
        path=None: str or None
            path to bibfile
        bibdb=None: bibtexparser.bibdatabase.BibDatabase or None
        encoding="utf8": str
            bib file encoding
        skip_ioerrors: bool
            if False, an IOError will be raised if
            newglossaryterm or newacronym is badly formatted
        ignore_unknown_types: bool
            if True, strip unknown types, otherwise raise a ValueError

        Notes
        -----
        the texsoup package is required.

        if a newglossaryterm has field 'type={symbols}', then
        it will be loaded as a symbol

        """
        from ipypublish.bib2glossary.parse_tex import parse_tex

        gterms, acronyms = parse_tex(
            text_str=text_str, path=path, encoding=encoding, skip_ioerrors=skip_ioerrors
        )
        entries = {}
        for key, fields in gterms.items():

            fields["ENTRYTYPE"] = ETYPE_GLOSS
            if fields.get("type", None) == "symbols":
                fields["ENTRYTYPE"] = ETYPE_SYMBOL
                fields.pop("type")
            elif "type" in fields:
                if not ignore_unknown_types:
                    raise ValueError(
                        "the 'type' is not recognised: " "{}".format(fields["type"])
                    )
                fields.pop("type")

            fields["ID"] = key
            entry = BibGlossEntry(fields)
            entries[entry.key] = entry

        for key, fields in acronyms.items():
            fields["ENTRYTYPE"] = ETYPE_ACRONYM
            fields["ID"] = key
            entry = BibGlossEntry(fields)
            entries[entry.key] = entry

        self._entries = entries

        return True

    @staticmethod
    def guess_path(path):
        """ guess the path of a bib file, with or without a file extension,
        from the available files in the path folder
        """
        basepath, extension = os.path.splitext(str(path))
        if extension in [".bib", ".biblatex", ".bibtex"]:
            return path
        elif extension in [".tex", ".latex"]:
            return path
        elif os.path.exists(basepath + ".bib"):
            return basepath + ".bib"
        elif os.path.exists(basepath + ".bibtex"):
            return basepath + ".bibtex"
        elif os.path.exists(basepath + ".biblatex"):
            return basepath + ".biblatex"
        elif os.path.exists(basepath + ".tex"):
            return basepath + ".tex"
        elif os.path.exists(basepath + ".latex"):
            return basepath + ".latex"
        else:
            return None

    def load(self, path, encoding="utf8"):
        """load a file, the type will be guessed from the extension,
        or (if no extension is given), the available files in the path folder

        Parameters
        ----------
        path: str
        encoding='utf8': str
            encoding of the file

        """
        path = self.guess_path(path)
        if path is None:
            raise IOError("no acceptable loader found for path: {}".format(path))
        basepath, extension = os.path.splitext(str(path))
        if extension in [".bib", ".biblatex", ".bibtex"]:
            self.load_bib(path=path, encoding=encoding)
        elif extension in [".tex", ".latex"]:
            self.load_tex(path=path, encoding=encoding)

    def to_dict(self):
        return {k: e.to_dict() for k, e in self.items()}

    def to_bib_string(self):
        bibdb = bibtexparser.bibdatabase.BibDatabase()
        bibdb.entries = [e.to_dict() for e in self.values()]
        writer = bibtexparser.bwriter.BibTexWriter()
        writer.contents = ["comments", "entries"]
        writer.indent = "  "
        # writer.order_entries_by = ('ENTRYTYPE', 'ID')
        return writer.write(bibdb)

    def to_latex_dict(self, splitlines=True):
        """convert to dict of latex strings

        Returns
        -------
        dict:
            {(<type>, <key>): <latex string>}

        """
        latex_stings = {}
        for entry in self.values():
            string = entry.to_latex()
            if splitlines:
                string = string.splitlines()
            latex_stings[(entry.type, entry.key)] = string
        return latex_stings

    def to_latex_string(self):
        lines = []
        latex_dict = self.to_latex_dict(splitlines=False)
        for key in sorted(list(latex_dict.keys())):
            lines.append(latex_dict[key])
        return "\n".join(lines)



import pytest


@pytest.mark.requires_latexmk
@pytest.mark.ipynb("nb_complex_outputs")
def test_complex_latex(ipynb_app):
    """ includes:

    - international language (portugese)
    - internal (image) files
    - external logo and bib

    TODO non-ascii characters in text not working on Linux
    This works locally (on osx) and on travis osx, but not for linux
    latexmk raises `Package utf8x Error: MalformedUTF-8sequence.` or
    `character used is undefined`
    presumable this is due to the way the ipynb is being encoded on osx

    """
    ipynb_app.run(
        {
            "conversion": "latex_ipypublish_main",
            "default_pporder_kwargs": {"create_pdf": True},
            "default_ppconfig_kwargs": {"pdf_debug": True},
        }
    )

    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_exists(extension=".pdf")
    ipynb_app.assert_converted_equals_expected("latex_ipypublish_main")


@pytest.mark.ipynb("nb_complex_outputs")
def test_complex_html(ipynb_app):
    """ includes:

    - internal (image) files
    - external logo and bib

    """
    ipynb_app.run({"conversion": "html_ipypublish_main"})

    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_equals_expected("html_ipypublish_main")


@pytest.mark.ipynb("nb_complex_outputs")
def test_complex_slides(ipynb_app):
    """ includes:

    - internal (image) files
    - external logo and bib

    """
    ipynb_app.run(
        {
            "conversion": "slides_ipypublish_main",
            "default_pporder_kwargs": {"slides": True},
        }
    )

    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_equals_expected("slides_ipypublish_main")


@pytest.mark.ipynb("nb_complex_outputs")
def test_complex_sphinx_rst(ipynb_app):
    """ includes:

    - internal (image) files
    - external logo and bib

    """
    if ipynb_app.pandoc_version < "2.6":
        raise AssertionError("pandoc version must be >= 2.6")

    ipynb_app.run(
        {
            "conversion": "sphinx_ipypublish_all",
            "default_pporder_kwargs": {"slides": True},
        }
    )

    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_equals_expected("sphinx_ipypublish_all.pandoc.2-6")



import sys
import os
import argparse
import fnmatch

from ipypublish import __version__
from ipypublish.convert.config_manager import iter_all_export_infos


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    pass


class CustomParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n" % message)
        self.print_help()
        sys.exit(2)


def get_parser(**kwargs):
    return CustomParser(formatter_class=CustomFormatter, **kwargs)


def get_plugin_str(plugin_folder_paths, regex, verbose):
    """return string listing all available export configurations """
    outstrs = []
    # outstrs.append('Available Export Configurations')
    # outstrs.append('-------------------------------')
    configs = [
        e
        for e in iter_all_export_infos(plugin_folder_paths, get_mime=verbose)
        if fnmatch.fnmatch(e["key"], "*{}*".format(regex))
    ]

    for item in sorted(configs, key=lambda i: (i["class"], i["key"])):
        outstrs.append("- Key:   {}".format(item["key"]))
        outstrs.append("  Class: {}".format(item["class"]))
        path = item["path"].split(os.path.sep)
        if verbose:
            outstrs.append("  Type:  {}".format(item["mime_type"]))
            path = os.path.join(*path)
        else:
            path = os.path.join("...", *path[-3:])

        if len(path) < 4:
            outstrs.append("  Path:  {}".format(item["path"]))
        else:
            outstrs.append("  Path:  {}".format(path))

        outstrs.append("  About: {}".format(item["description"][0].strip()))
        if verbose:
            for descript in item["description"][1:]:
                outstrs.append("         {}".format(descript.strip()))
        # note could wrap description (less than x characters)
        outstrs.append(" ")

    return "\n".join(outstrs)


def parse_options(sys_args, program):

    if program not in ["nbpublish", "nbpresent"]:
        raise ValueError("program should be nbpublish or nbpresent")

    if program == "nbpresent":
        parser = get_parser(
            description=(
                "load reveal.js slides as a web server,\n"
                "converting from ipynb first "
                "if path extension is `ipynb`"
            )
        )
        file_help = "path to html or ipynb file"
        default_key = "slides_ipypublish_main"
    else:
        parser = get_parser(
            description=(
                "convert one or more Jupyter notebooks " "to a publishable format"
            )
        )
        file_help = "notebook file or directory"
        default_key = "latex_ipypublish_main"

    parser.add_argument("--version", action="version", version=__version__)

    parser.add_argument(
        "filepath", type=str, nargs="?", help=file_help, metavar="filepath"
    )

    parser.add_argument(
        "-f",
        "--outformat",
        type=str,
        metavar="key|filepath",
        help=(
            "export format configuration to use, "
            "can be a key name or path to the file"
        ),
        default=default_key,
    )

    export_group = parser.add_argument_group("export configurations")
    export_group.add_argument(
        "-ep",
        "--export-paths",
        action="append",
        metavar="path",
        type=str,
        help=("add additional folder paths, " "containing export configurations"),
        default=[],
    )
    export_group.add_argument(
        "-le",
        "--list-exporters",
        type=str,
        metavar="filter",
        nargs="?",
        const="*",
        help=("list export configurations, " "optionally filtered e.g. -le html*"),
    )
    export_group.add_argument(
        "-lv",
        "--list-verbose",
        action="store_true",
        help=("when listing export configurations, " "give a verbose description"),
    )

    nbmerge_group = parser.add_argument_group("nb merge")
    nbmerge_group.add_argument(
        "-i",
        "--ignore-prefix",
        type=str,
        metavar="str",
        default="_",
        help="ignore ipynb files with this prefix",
    )

    output_group = parser.add_argument_group("output")
    output_group.add_argument(
        "-o",
        "--outpath",
        type=str,
        metavar="str",
        help="path to output converted files",
        default=os.path.join(os.getcwd(), "converted"),
    )
    # output_group.add_argument("-d","--dump-files", action="store_true",
    #                     help='dump external files, '
    #                          'linked to in the document, into the outpath')
    output_group.add_argument(
        "-c",
        "--clear-files",
        action="store_true",
        help=("clear any external files " "that already exist in the outpath"),
    )

    if program == "nbpublish":
        pdf_group = parser.add_argument_group("pdf export")
        pdf_group.add_argument(
            "-pdf",
            "--create-pdf",
            action="store_true",
            help="convert to pdf (only if latex exporter)",
        )
        pdf_group.add_argument(
            "-ptemp",
            "--pdf-in-temp",
            action="store_true",
            help=(
                "run pdf conversion in a temporary folder"
                " and only copy back the .pdf file"
            ),
        )
        pdf_group.add_argument(
            "-pbug",
            "--pdf-debug",
            action="store_true",
            help="run latexmk in interactive mode",
        )

        view_group = parser.add_argument_group("view output")
        view_group.add_argument(
            "-lb",
            "--launch-browser",
            action="store_true",
            help="open the output in an available web-browser",
        )

    debug_group = parser.add_argument_group("debugging")
    debug_group.add_argument(
        "-log",
        "--log-level",
        type=str,
        default="info",
        choices=["debug", "info", "warning", "error"],
        help="the logging level to output to screen/file",
    )
    debug_group.add_argument(
        "-pt",
        "--print-traceback",
        action="store_true",
        help=("print the full exception traceback"),
    )
    debug_group.add_argument(
        "-dr",
        "--dry-run",
        action="store_true",
        help=("perform a 'dry run', " "which will not output any files"),
    )

    args = parser.parse_args(sys_args)
    options = vars(args)

    filepath = options.pop("filepath")
    list_plugins = options.pop("list_exporters")
    list_verbose = options.pop("list_verbose")

    if filepath is None and list_plugins:
        parser.exit(
            message=get_plugin_str(options["export_paths"], list_plugins, list_verbose)
        )
    elif filepath is None:
        parser.error("no filepath specified")

    return filepath, options



# -*- coding: utf-8 -*-
"""
caching of the bibglossaries

Originally adapted from: https://github.com/mcmtroffaes/sphinxcontrib-bibtex
"""

import six

try:  # pragma: no cover
    from collections import OrderedDict
except ImportError:  # pragma: no cover
    from ordereddict import OrderedDict
import ast
import collections
import copy
from ordered_set import OrderedSet
import re


def _raise_invalid_node(node):
    """Helper method to raise an exception when an invalid node is
    visited.
    """
    raise ValueError("invalid node %s in filter expression" % node)


class _FilterVisitor(ast.NodeVisitor):

    """Visit the abstract syntax tree of a parsed filter expression."""

    entry = None
    """The bibliographic entry to which the filter must be applied."""

    cited_docnames = False
    """The documents where the entry is cited (empty if not cited)."""

    def __init__(self, entry, docname, cited_docnames):
        self.entry = entry
        self.docname = docname
        self.cited_docnames = cited_docnames

    def visit_Module(self, node):  # noqa: N802
        if len(node.body) != 1:
            raise ValueError("filter expression cannot contain multiple expressions")
        return self.visit(node.body[0])

    def visit_Expr(self, node):  # noqa: N802
        return self.visit(node.value)

    def visit_BoolOp(self, node):  # noqa: N802
        outcomes = (self.visit(value) for value in node.values)
        if isinstance(node.op, ast.And):
            return all(outcomes)
        elif isinstance(node.op, ast.Or):
            return any(outcomes)
        else:  # pragma: no cover
            # there are no other boolean operators
            # so this code should never execute
            assert False, "unexpected boolean operator %s" % node.op

    def visit_UnaryOp(self, node):  # noqa: N802
        if isinstance(node.op, ast.Not):
            return not self.visit(node.operand)
        else:
            _raise_invalid_node(node)

    def visit_BinOp(self, node):  # noqa: N802
        left = self.visit(node.left)
        op = node.op
        right = self.visit(node.right)
        if isinstance(op, ast.Mod):
            # modulo operator is used for regular expression matching
            if not isinstance(left, six.string_types):
                raise ValueError("expected a string on left side of %s" % node.op)
            if not isinstance(right, six.string_types):
                raise ValueError("expected a string on right side of %s" % node.op)
            return re.search(right, left, re.IGNORECASE)
        elif isinstance(op, ast.BitOr):
            return left | right
        elif isinstance(op, ast.BitAnd):
            return left & right
        else:
            _raise_invalid_node(node)

    def visit_Compare(self, node):  # noqa: N802
        # keep it simple: binary comparators only
        if len(node.ops) != 1:
            raise ValueError("syntax for multiple comparators not supported")
        left = self.visit(node.left)
        op = node.ops[0]
        right = self.visit(node.comparators[0])
        if isinstance(op, ast.Eq):
            return left == right
        elif isinstance(op, ast.NotEq):
            return left != right
        elif isinstance(op, ast.Lt):
            return left < right
        elif isinstance(op, ast.LtE):
            return left <= right
        elif isinstance(op, ast.Gt):
            return left > right
        elif isinstance(op, ast.GtE):
            return left >= right
        elif isinstance(op, ast.In):
            return left in right
        elif isinstance(op, ast.NotIn):
            return left not in right
        else:
            # not used currently: ast.Is | ast.IsNot
            _raise_invalid_node(op)

    def visit_Name(self, node):  # noqa: N802
        """Calculate the value of the given identifier."""
        id_ = node.id
        if id_ == "type":
            return self.entry.type.lower()
        elif id_ == "key":
            return self.entry.key.lower()
        elif id_ == "cited":
            return bool(self.cited_docnames)
        elif id_ == "docname":
            return self.docname
        elif id_ == "docnames":
            return self.cited_docnames
        elif id_ == "True":
            return True
        elif id_ == "False":
            return False
        else:
            return self.entry.get(id_, "")

    def visit_Set(self, node):  # noqa: N802
        return frozenset(self.visit(elt) for elt in node.elts)

    def visit_Str(self, node):  # noqa: N802
        return node.s

    # NameConstant is Python 3.4 only so do not insist on coverage
    def visit_NameConstant(self, node):  # noqa: N802
        return node.value

    def generic_visit(self, node):
        _raise_invalid_node(node)


class Cache:

    """Global bibgloss extension information cache. Stored in
    ``app.env.bibgloss_cache``, so must be picklable.
    """

    bibfiles = None
    """A :class:`dict` mapping .bib file names (relative to the top
    source folder) to :class:`BibfileCache` instances.
    """

    _bibliographies = None
    """Each bibglossary directive is assigned an id of the form
    bibtex-bibglossary-xxx. This :class:`dict` maps each docname
    to another :class:`dict` which maps each id
    to information about the bibliography directive,
    :class:`BibliographyCache`. We need to store this extra
    information separately because it cannot be stored in the
    :class:`~ipypublish.sphinx.gls.nodes.BibGlossaryNode` nodes
    themselves.
    """

    _cited = None
    """A :class:`dict` mapping each docname to a :class:`set` of
    citation keys.
    """

    _enum_count = None
    """A :class:`dict` mapping each docname to an :class:`int`
    representing the current bibliography enumeration counter.
    """

    def __init__(self):

        self.bibfiles = {}
        self._bibliographies = collections.defaultdict(dict)
        self._cited = collections.defaultdict(OrderedSet)
        self._enum_count = {}

    def purge(self, docname):
        """Remove  all information related to *docname*.

        :param docname: The document name.
        :type docname: :class:`str`
        """
        self._bibliographies.pop(docname, None)
        self._cited.pop(docname, None)
        self._enum_count.pop(docname, None)

    def inc_enum_count(self, docname):
        """Increment enumeration list counter for document *docname*."""
        self._enum_count[docname] += 1

    def set_enum_count(self, docname, value):
        """Set enumeration list counter for document *docname* to *value*."""
        self._enum_count[docname] = value

    def get_enum_count(self, docname):
        """Get enumeration list counter for document *docname*."""
        return self._enum_count[docname]

    def add_cited(self, key, docname):
        """Add the given *key* to the set of cited keys for
        *docname*.

        :param key: The citation key.
        :type key: :class:`str`
        :param docname: The document name.
        :type docname: :class:`str`
        """
        self._cited[docname].add(key)

    def get_cited_docnames(self, key):
        """Return the *docnames* from which the given *key* is cited.

        :param key: The citation key.
        :type key: :class:`str`
        """
        return frozenset(
            [docname for docname, keys in six.iteritems(self._cited) if key in keys]
        )

    def get_label_from_key(self, key):
        """Return label for the given key."""
        for bibcache in self.get_all_bibliography_caches():
            if key in bibcache.labels:
                return bibcache.labels[key]
        else:
            raise KeyError("%s not found" % key)

    def get_plural_from_key(self, key):
        """Return label for the given key."""
        for bibcache in self.get_all_bibliography_caches():
            if key in bibcache.plurals:
                return bibcache.plurals[key]
        else:
            raise KeyError("%s not found" % key)

    def get_all_cited_keys(self):
        """Yield all citation keys, sorted first by document
        (alphabetical), then by citation order in the document.
        """
        for docname in sorted(self._cited):
            for key in self._cited[docname]:
                yield key

    def set_bibliography_cache(self, docname, id_, bibcache):
        """Register *bibcache* (:class:`BibliographyCache`)
        with id *id_* for document *docname*.
        """
        assert id_ not in self._bibliographies[docname]
        self._bibliographies[docname][id_] = bibcache

    def get_bibliography_cache(self, docname, id_):
        """Return :class:`BibliographyCache` with id *id_* in
        document *docname*.
        """
        return self._bibliographies[docname][id_]

    def get_all_bibliography_caches(self):
        """Return all bibliography caches."""
        for bibcaches in six.itervalues(self._bibliographies):
            for bibcache in six.itervalues(bibcaches):
                yield bibcache

    def _get_bibliography_entries(self, docname, id_, warn):
        """Return filtered bibliography entries, sorted by occurence
        in the bib file.
        """
        # get the information of this bibliography node
        bibcache = self.get_bibliography_cache(docname=docname, id_=id_)
        # generate entries
        for bibfile in bibcache.bibfiles:
            for entry in self.bibfiles[bibfile].data.values():
                # beware: the prefix is not stored in the data
                # to allow reusing the data for multiple bibliographies
                cited_docnames = self.get_cited_docnames(bibcache.keyprefix + entry.key)
                visitor = _FilterVisitor(
                    entry=entry, docname=docname, cited_docnames=cited_docnames
                )
                try:
                    success = visitor.visit(bibcache.filter_)
                except ValueError as err:
                    warn("syntax error in :filter: expression; %s" % err)
                    # recover by falling back to the default
                    success = bool(cited_docnames)
                if success:
                    # entries are modified in an unpickable way
                    # when formatting, so fetch a deep copy
                    # and return this copy with prefixed key
                    # we do not deep copy entry.collection because that
                    # consumes enormous amounts of memory
                    # entry.collection = None
                    entry2 = copy.deepcopy(entry)
                    entry2.key = bibcache.keyprefix + entry.key
                    yield entry2

    def get_bibliography_entries(self, docname, id_, warn):
        """Return filtered bibliography entries, sorted by citation order."""
        # get entries, ordered by bib file occurrence
        entries = OrderedDict(
            (entry.key, entry)
            for entry in self._get_bibliography_entries(
                docname=docname, id_=id_, warn=warn
            )
        )
        # order entries according to which were cited first
        # first, we add all keys that were cited
        # then, we add all remaining keys
        sorted_entries = []
        for key in self.get_all_cited_keys():
            try:
                entry = entries.pop(key)
            except KeyError:
                pass
            else:
                sorted_entries.append(entry)
        sorted_entries += six.itervalues(entries)
        return sorted_entries


class BibfileCache(collections.namedtuple("BibfileCache", "mtime data")):

    """Contains information about a parsed .bib file.

    .. attribute:: mtime

        A :class:`float` representing the modification time of the .bib
        file when it was last parsed.

    .. attribute:: data

        A ipypublish.bib2glossary.BibGlossDB instance

    """


class BibliographyCache(
    collections.namedtuple(
        "BibliographyCache",
        """bibfiles encoding style unsorted labels plurals filter_ keyprefix
""",
    )
):

    """Contains information about a bibliography directive.

    .. attribute:: bibfiles

        A :class:`list` of :class:`str`\\ s containing the .bib file
        names (relative to the top source folder) that contain the
        references.

    .. attribute:: encoding

        The encoding of the glossary file.

    .. attribute:: style

        The glossary style.

    .. attribute:: unsorted

        If True the glossary terms will be sorted by order of use,
        rather than alphabetically

    .. attribute:: labels

        Maps citation keys to their final labels.

    .. attribute:: plurals

        Maps citation keys to their final pluralised labels.

    .. attribute:: keyprefix

        This bibliography's string prefix for citation keys.

    .. attribute:: filter_

        An :class:`ast.AST` node, containing the parsed filter expression.
    """



import ruamel.yaml as yaml
from nbformat.notebooknode import NotebookNode


def recurse_convert(node):
    # type: (NotebookNode) -> dict
    """convert notebook node to dict"""
    dct = {}
    for key, val in node.items():
        if isinstance(val, NotebookNode):
            dct[key] = recurse_convert(val)
        else:
            dct[key] = val
    return dct


def meta_to_yaml(metadata, comment="#~~ "):
    # type: (NotebookNode, str) -> str
    """convert metadata json to yaml

    Parameters
    ----------
    metadata: nbformat.notebooknode.NotebookNode
    comment="": str
        append to the start of each line

    Returns
    -------
    string: str

    """
    metadata = recurse_convert(metadata)
    string = yaml.dump(metadata, default_flow_style=False)
    if comment:
        string = "\n".join([comment + s for s in string.splitlines()])
    return string



""" a module to convert between the old (Python script) plugin format,
and the new (JSON) one
"""
from typing import Dict, Tuple  # noqa: F401
import ast
import json


def assess_syntax(path):

    with open(path) as file_obj:
        content = file_obj.read()

    syntax_tree = ast.parse(content)

    docstring = ""  # docstring = ast.get_docstring(syntaxTree)
    unknowns = []
    imported = {}
    assignments = {}
    for i, child in enumerate(ast.iter_child_nodes(syntax_tree)):
        if i == 0 and isinstance(child, ast.Expr) and isinstance(child.value, ast.Str):
            docstring = child.value.s
        elif isinstance(child, ast.ImportFrom):
            module = child.module
            for n in child.names:
                import_pth = module + "." + n.name
                imported[n.name if n.asname is None else n.asname] = import_pth
        elif isinstance(child, ast.Assign):
            targets = child.targets
            if len(targets) > 1:
                raise IOError(
                    "cannot handle expansion assignments " "(e.g. `a, b = [1, 2]`)"
                )
            target = child.targets[0]  # type: ast.Name
            assignments[target.id] = child.value
        else:
            unknowns.append(child)

    if unknowns:
        print(
            "Warning this script can only handle 'ImportFrom' and 'Assign' "
            "syntax, found additional items: {}".format(unknowns)
        )

    return docstring, imported, assignments


def ast_to_json(item, imported, assignments):
    """recursively convert ast items to json friendly values"""
    value = None
    if item in ["True", "False", "None"]:  # python 2.7
        value = {"True": True, "False": False, "None": None}[item]
    elif hasattr(ast, "NameConstant") and isinstance(item, ast.NameConstant):
        value = item.value
    elif isinstance(item, ast.Str):
        value = item.s
    elif isinstance(item, ast.Num):
        value = item.n
    elif isinstance(item, ast.Name):
        if item.id in imported:
            value = imported[item.id]
        elif item.id in assignments:
            value = ast_to_json(assignments[item.id], imported, assignments)
        elif item.id in ["True", "False", "None"]:  # python 2.7
            value = {"True": True, "False": False, "None": None}[item.id]
        else:
            raise ValueError("could not find assignment '{}' in config".format(item.id))
    elif isinstance(item, (ast.List, ast.Tuple, ast.Set)):
        value = [ast_to_json(i, imported, assignments) for i in item.elts]
    elif isinstance(item, ast.Dict):
        value = convert_dict(item, imported, assignments)
    else:
        raise ValueError("could not handle ast item: {}".format(item))

    return value


def convert_dict(dct, imported, assignments):
    # type: (ast.Dict, Dict[str, str], dict) -> dict
    """recurse through and replace keys"""
    out_dict = {}
    for key, val in zip(dct.keys, dct.values):
        if not isinstance(key, ast.Str):
            raise ValueError("expected key to be a Str; {}".format(key))
        out_dict[key.s] = ast_to_json(val, imported, assignments)

    return out_dict


def convert_oformat(oformat):

    if oformat == "Notebook":
        outline = None  # TODO do notebooks need template (they have currently)
        exporter = "nbconvert.exporters.NotebookExporter"
    elif oformat == "Latex":
        exporter = "nbconvert.exporters.LatexExporter"
        outline = {
            "module": "ipypublish.templates.outline_schemas",
            "file": "latex_outline.latex.j2",
        }
    elif oformat == "HTML":
        exporter = "nbconvert.exporters.HTMLExporter"
        outline = {
            "module": "ipypublish.templates.outline_schemas",
            "file": "html_outline.html.j2",
        }
    elif oformat == "Slides":
        exporter = "nbconvert.exporters.SlidesExporter"
        outline = {
            "module": "ipypublish.templates.outline_schemas",
            "file": "html_outline.html.j2",
        }
    else:
        raise ValueError(
            "expected oformat to be: " "'Notebook', 'Latex', 'HTML' or 'Slides'"
        )
    return exporter, outline


def convert_config(config, exporter_class, allow_other):
    # type: (dict, str) -> dict
    """convert config into required exporter format"""
    filters = {}
    preprocs = {}
    other = {}
    # first parse
    for key, val in config.items():
        # TODO Exporter.filters and TemplateExporter.filters always the same?
        if key in ["Exporter.filters", "TemplateExporter.filters"]:
            filters.update(config[key])
        if key in ["Exporter.preprocessors", "TemplateExporter.preprocessors"]:
            if preprocs:
                raise ValueError(
                    "'config' contains both Exporter.preprocessors and "
                    "TemplateExporter.preprocessors"
                )
            for p in val:
                pname = p.split(".")[-1]
                preprocs[pname] = {"class": p, "args": {}}
                # TODO move these special cases to seperate input/function
                if pname in ["LatexDocLinks", "LatexDocHTML"]:
                    preprocs[pname]["args"]["metapath"] = "${meta_path}"
                    preprocs[pname]["args"]["filesfolder"] = "${files_path}"

    # second parse
    for key, val in config.items():
        if key in [
            "Exporter.filters",
            "TemplateExporter.filters",
            "Exporter.preprocessors",
            "TemplateExporter.preprocessors",
        ]:
            continue
        if key.split(".")[0] in preprocs:
            preprocs[key.split(".")[0]]["args"][".".join(key.split(".")[1:])] = val
        else:
            other[key] = val

    if other and not allow_other:
        print("Warning: ignoring other args: {}".format(other))
        other = {}

    output = {
        "class": exporter_class,
        "filters": filters,
        "preprocessors": list(preprocs.values()),
        "other_args": other,
    }
    return output


def replace_template_path(path):
    """ replace original template path with new dict """
    segments = path.split(".")
    module = ".".join(segments[0:-1])
    name = segments[-1]
    if module == "ipypublish.html.ipypublish":
        return {
            "module": "ipypublish.templates.segments",
            "file": "ipy-{0}.html-tplx.json".format(name),
        }
    elif module == "ipypublish.html.standard":
        return {
            "module": "ipypublish.templates.segments",
            "file": "std-{0}.html-tplx.json".format(name),
        }
    elif module == "ipypublish.latex.standard":
        return {
            "module": "ipypublish.templates.segments",
            "file": "std-{0}.latex-tpl.json".format(name),
        }
    elif module == "ipypublish.latex.ipypublish":
        return {
            "module": "ipypublish.templates.segments",
            "file": "ipy-{0}.latex-tpl.json".format(name),
        }
    else:
        print("Warning: unknown template path: {}".format(path))
        return {"module": module, "file": "{0}.json".format(name)}


def create_json(docstring, imported, assignments, allow_other=True):
    #  type: (str, Dict[str, str], dict, bool) -> dict
    """Set docstring here.

    Parameters
    ----------
    docstring: str
        the doc string of the module
    imported: dict
        imported classes
    assignments: dict
        assigned values (i.e. 'a = b')
    allow_other: bool
        whether to allow arguments in config,
        which do not relate to preprocessors

    Returns
    -------

    """

    oformat = None
    config = None
    template = None
    for value, expr in assignments.items():
        if value == "oformat":
            if not isinstance(expr, ast.Str):
                raise ValueError("expected 'oformat' to be a Str; {}".format(expr))
            oformat = expr.s
        elif value == "config":
            if not isinstance(expr, ast.Dict):
                raise ValueError("expected 'config' to be a Dict; {}".format(expr))
            config = convert_dict(expr, imported, assignments)
        elif value == "template":
            if not isinstance(expr, ast.Call):
                raise ValueError("expected 'config' to be a call to create_tpl(x)")
            # func = expr.func  # TODO make sure func name is create_tpl/tplx
            args = expr.args
            keywords = expr.keywords
            if len(args) != 1 or len(keywords) > 0:
                raise ValueError("expected create_tpl(x) to have one argument")
            seg_list = args[0]
            if isinstance(seg_list, ast.ListComp):
                seg_list = seg_list.generators[0].iter
            if not isinstance(seg_list, ast.List):
                raise ValueError(
                    "expected create_tpl(x) arg to be a List; {}".format(seg_list)
                )
            segments = []
            for seg in seg_list.elts:
                if isinstance(seg, ast.Attribute):
                    seg_name = seg.value.id
                elif isinstance(seg, ast.Name):
                    seg_name = seg.id
                else:
                    raise ValueError(
                        "expected seg in template to be an Attribute; "
                        + "{1}".format(seg)
                    )

                if seg_name not in imported:
                    raise ValueError("segment '{}' not found".format(seg_name))
                segments.append(imported[seg_name])
            template = segments

    if oformat is None:
        raise ValueError("could not find 'oformat' assignment")
    if config is None:
        raise ValueError("could not find 'config' assignment")
    if template is None:
        raise ValueError("could not find 'template' assignment")

    exporter_class, outline = convert_oformat(oformat)
    exporter = convert_config(config, exporter_class, allow_other)

    if any(["biblio_natbib" in s for s in template]):
        exporter["filters"]["strip_ext"] = "ipypublish.filters.filters.strip_ext"

    return {
        "description": docstring.splitlines(),
        "exporter": exporter,
        "template": None
        if outline is None
        else {
            "outline": outline,
            "segments": [replace_template_path(s) for s in template],
        },
    }


def convert_to_json(path, outpath=None, ignore_other=False):
    """Set docstring here.

    Parameters
    ----------
    path: str
        input module path
    outpath=None: str or None
        if set, output json to this path
    ignore_other: bool
        whether to ignore arguments in config,
        which do not relate to preprocessors
    Returns
    -------

    """
    _docstring, _imported, _assignments = assess_syntax(path)
    # print(_docstring)
    # print()
    # print(_imported)
    # print()
    # print(_assignments)
    output = create_json(_docstring, _imported, _assignments, not ignore_other)
    if outpath:
        with open(outpath, "w") as file_obj:
            json.dump(output, file_obj, indent=2)
    return json.dumps(output, indent=2)


if __name__ == "__main__":

    if False:
        import glob
        import os

        for path in glob.glob(
            "/Users/cjs14/GitHub/ipypublish" "/ipypublish/export_plugins/*.py"
        ):
            dirname = os.path.dirname(path)
            name = os.path.splitext(os.path.basename(path))[0]
            try:
                convert_to_json(
                    path, os.path.join(dirname, name + ".json"), ignore_other=True
                )
            except ValueError as err:
                print("{0} failed: {1}".format(path, err))

    convert_to_json(
        "/Users/cjs14/GitHub/ipypublish" "/ipypublish_plugins/example_new_plugin.py",
        "/Users/cjs14/GitHub/ipypublish" "/ipypublish_plugins/example_new_plugin.json",
    )



import os
import pytest


@pytest.mark.requires_latexmk
@pytest.mark.ipynb("nb_markdown_cells")
def test_latex_and_pdf(ipynb_app):
    ipynb_app.run(
        {
            "conversion": "latex_ipypublish_main",
            "default_pporder_kwargs": {"create_pdf": True},
        }
    )
    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_exists(extension=".pdf")
    # Table format has changed through the versions
    if ipynb_app.pandoc_version < "2.2":
        raise AssertionError("pandoc version must be >= 2.2")
    ipynb_app.assert_converted_equals_expected("latex_ipypublish_main.pandoc.2-2")


@pytest.mark.skipif(os.name == "nt", reason="skipping")
@pytest.mark.ipynb("nb_markdown_cells")
def test_sphinx_rst(ipynb_app):
    ipynb_app.run({"conversion": "sphinx_ipypublish_main"})
    ipynb_app.assert_converted_exists()
    # Table format has changed through the versions
    if ipynb_app.pandoc_version < "2.6":
        raise AssertionError("pandoc version must be >= 2.6")
    ipynb_app.assert_converted_equals_expected("sphinx_ipypublish_main.pandoc.2-6")


@pytest.mark.skipif(os.name == "nt", reason="skipping")
@pytest.mark.ipynb("nb_with_mkdown_images")  # out_to_temp=False
def test_sphinx_rst_with_mkdown_images(ipynb_app):
    """ test a notebook with multiple images """
    ipynb_app.run(
        {
            "conversion": "sphinx_ipypublish_main.run",
            "log_to_file": True,
            "default_pporder_kwargs": {"dump_files": True},
        }
    )
    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_equals_expected("sphinx_ipypublish_main")
    assert ipynb_app.converted_path.joinpath("main_files/example.jpg").is_file()



from docutils import nodes


class BibGlossaryNode(nodes.General, nodes.Element):

    """Node for representing a bibglossary. Replaced by a list of
    glossary terms by
    :class:`~ipypublish.sphinx.gls.transforms.BibGlossaryTransform`.
    """

    pass
